//Compact Solitaire
//Version 1.1 (November 2006)
//Copyright 2006 by Donnie Russell



#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "interface.h"



//...........................................................................
#define CARDW  70
#define CARDH  105

#define MAX_FOUNDATION_STACKS  30
#define MAX_TABLEAU_STACKS     30
#define MAX_CELL_STACKS        30

#define NUM_GAME_TYPES  10

#define MAX_UNDO  500

#define MAX_DECKS  2
#define MAX_CARDS  (13*4*MAX_DECKS)



struct CARD_STRUCT {char rank, suit, facedown; struct CARD_STRUCT *next;};
typedef struct CARD_STRUCT CARD;

typedef struct {CARD *card; int x, y; char type, showempty;} STACK;

typedef struct {char recycle; STACK *src_stack; CARD *card; STACK *dst_stack; int counter;} UNDO;
//...........................................................................



//...........................................................................
IMG           BackIMG = {0, 0, 0};
IMG          CourtIMG = {0, 0, 0};
IMG          CardsIMG = {0, 0, 0};
IMG FoundationRankIMG = {0, 0, 0};
IMG          SpadeIMG = {0, 0, 0};

char CardsRendered[13*4];

int FoundationRankRendered; //-2: nothing rendered yet

STACK Hand;
STACK Deck;
STACK Waste;
STACK Reserve;
STACK Foundation[MAX_FOUNDATION_STACKS];
STACK Tableau[MAX_TABLEAU_STACKS];
STACK Cell[MAX_CELL_STACKS];

int NumFoundationStacks;
int NumTableauStacks;
int NumCellStacks;

CARD Recycle; //this special card signals a recycle
CARD FlipLast; //this special card signals a flip of the last card in a stack

STACK *HandFromStack;
int HandOffsetX;
int HandOffsetY;

int GameType;

unsigned int GameSeed;

char *GameTitle[NUM_GAME_TYPES] =
{
  "Klondike",
  "FreeCell",
  "Canfield",
  "Golf",
  "Spider",
  "Pyramid",
  "Yukon",
  "Scorpion",
  "Monte Carlo",
  "Forty Thieves"
};

int MouseLatch;
int SynchroCounter;

int FoundationRank; //rank of starting Foundation card (-1 if not set)
int NumRecyclesLeft; //number of times Waste cards can be moved back to Deck (-1 if infinite)
int NumCardsDrawn; //max number of cards moved from Deck to Waste on each click
int UseWasteAsFoundation; //use Waste as Foundation
int DealToEachTableauFromDeck; //move one card from Deck to each Tableau stack instead of to Waste
int FillCellsBetweenDeckAndWaste; //cards moved from Deck to Waste pass through intermediate Cells
int UseThirteenElimination; //if one card is dropped on another card and the first card is a king or the rank values of both cards add up to 13, the appropriate card(s) is(are) moved to the first Foundation stack
int UseMatchingRankElimination; //if one card is dropped on another card that has the same rank, both cards are moved to the Waste stack
int ConsolidateTableauAndFillFromDeck; //group all filled Tableau stacks at beginning and then fill each empty stack with one card from Deck

char ForegroundMessage[128] = "";

int EasterEggCount = 0;
int EasterEggFlag = 0;



int NumUndo;
int UndoIndex;
int UndoCounter;

char UndoRecycleFlag; //this flag is a signal to AddUndoEntry() that NumRecyclesLeft has decreased

UNDO Undo[MAX_UNDO];



CARD Cards[MAX_CARDS];
char CardsUsed[MAX_CARDS];
//...........................................................................



//...........................................................................
extern IMG ClientIMG;

extern FNT EasterEggMsgFNT;
extern FNT RankFNT;
extern FNT SuitFNT;
extern FNT SuitPatternFNT;
extern FNT LargeSuitFNT;
extern FNT TextFNT;
extern FNT FixedFNT;
//...........................................................................



//****************************************************************************
//------------------------------------------------
//Mersenne Twister random number generator MT19937
//------------------------------------------------



unsigned int stateMT[624+1];
unsigned int *nextMT;
int leftMT = -1;



void seedMT(unsigned int seed)
{
  unsigned int x = (seed|1U)&0xFFFFFFFFU, *s = stateMT;
  int j;

  for (leftMT=0, *s++=x, j=624; --j; *s++=(x*=69069U)&0xFFFFFFFFU);
}



unsigned int reloadMT(void)
{
  unsigned int *p0 = stateMT, *p2 = stateMT+2, *pM = stateMT+397, s0, s1;
  int j;

  if (leftMT<-1) seedMT(4357U);
  leftMT = 624-1;
  nextMT = stateMT+1;
  for (s0 = stateMT[0], s1 = stateMT[1], j = 624-397+1; --j; s0 = s1, s1 = *p2++)
    *p0++ = *pM++^(((s0&0x80000000U)|(s1&0x7FFFFFFFU))>>1)^(((s1)&0x00000001U)?0x9908B0DFU:0U);
  for (pM = stateMT, j = 397; --j; s0 = s1, s1 = *p2++)
    *p0++ = *pM++^(((s0&0x80000000U)|(s1&0x7FFFFFFFU))>>1)^(((s1)&0x00000001U)?0x9908B0DFU:0U);
  s1 = stateMT[0];
  *p0 = *pM^(((s0&0x80000000U)|(s1&0x7FFFFFFFU))>>1)^(((s1)&0x00000001U)?0x9908B0DFU:0U);
  s1 ^= (s1>>11);
  s1 ^= (s1<<7)&0x9D2C5680U;
  s1 ^= (s1<<15)&0xEFC60000U;
  return (s1^(s1>>18));
}



unsigned int randomMT(void)
{
  unsigned int y;

  if (--leftMT<0) return (reloadMT());
  y = *nextMT++;
  y ^= (y>>11);
  y ^= (y<<7)&0x9D2C5680U;
  y ^= (y<<15)&0xEFC60000U;
  return (y^(y>>18));
}



unsigned int rangeMT(unsigned int lower, unsigned int upper)
{
  return (((randomMT()&0x0000FFFFU)*(upper-lower+0x00000001U))>>16)+lower;
}
//***************************************************************************



//****************************************************************************
//-----------------------------
//Card and foundation rendering
//-----------------------------



double CardData[2*10*10] =
{
  3,3,  0,0,  0,0,  0,0,  0,0,  0,0,  0,0,  0,0,  0,0,  0,0,
  3,1,  3,5,  0,0,  0,0,  0,0,  0,0,  0,0,  0,0,  0,0,  0,0,
  3,1,  3,3,  3,5,  0,0,  0,0,  0,0,  0,0,  0,0,  0,0,  0,0,
  2,1,  4,1,  2,5,  4,5,  0,0,  0,0,  0,0,  0,0,  0,0,  0,0,
  2,1,  4,1,  3,3,  2,5,  4,5,  0,0,  0,0,  0,0,  0,0,  0,0,
  2,1,  4,1,  2,3,  4,3,  2,5,  4,5,  0,0,  0,0,  0,0,  0,0,
  2,1,  4,1,  3,2,  2,3,  4,3,  2,5,  4,5,  0,0,  0,0,  0,0,
  2,1,  4,1,  2,2.5,  4,2.5,  2,3.5,  4,3.5,  2,5,  4,5,  0,0,  0,0,
  2,1,  4,1,  2,2.25,  4,2.25,  3,3,  2,3.75,  4,3.75,  2,5,  4,5,  0,0,
  2,1,  4,1,  3,1.75,  2,2.5,  4,2.5,  2,3.5,  4,3.5,  3,4.25,  2,5,  4,5,
};

char *RankString[13] = {"A", "2", "3", "4", "5", "6", "7", "8", "9", "10", "J", "Q", "K"};

int SuitChar[4] = {0xa7, 0xa8, 0xa9, 0xaa}; //club, diamond, heart, spade

PIXEL_TYPE SuitColor[4] = {COLOR(0,0,0), COLOR(200,0,0), COLOR(200,0,0), COLOR(0,0,0)};



void RenderCard(int rank, int suit, int cx, int cy)
{
  IMG aIMG = {0, 0, 0};
  IMG bIMG = {0, 0, 0};

  int index, adjustx, adjusty, w, h, x, y, i;
  double dx, dy;
  PIXEL_TYPE border_color;

  char *easter_egg_message = "Donnie Russell";



  index = 13*suit+rank;

  if (CardsRendered[index]) goto render;



  if (AllocIMG(&aIMG, CARDW*2, CARDH*2)==0) goto cleanup;
  if (AllocIMG(&bIMG, CARDW*1, CARDH*1)==0) goto cleanup;



  //background
  if (EasterEggFlag==0)
  {
    for (i=0; i<15; i++) PutFilledRect(&aIMG, 0, i*(aIMG.h/15), aIMG.w-1, (i+1)*(aIMG.h/15)-1, COLOR(250, 250, 250));

    border_color = BLACK;
  }
  else
  {
    for (i=0; i<15; i++) PutFilledRect(&aIMG, 0, i*(aIMG.h/15), aIMG.w-1, (i+1)*(aIMG.h/15)-1, COLOR(128, 128, 255-i*3));

    //put easter egg message on ace of spades
    if (rank==0 && suit==3)
    {
      GetStringSizeFNT(&EasterEggMsgFNT, easter_egg_message, &w, &h);
      DrawStringFNT(&EasterEggMsgFNT, easter_egg_message, &aIMG, aIMG.w-w-4, 0, BLACK);
      DrawStringRevFlipFNT(&EasterEggMsgFNT, easter_egg_message, &aIMG, 4, aIMG.h-h, BLACK);
    }

    border_color = WHITE;
  }



  //border
  PutFilledRect(&aIMG, 0, 0, aIMG.w-1, 0, border_color);
  PutFilledRect(&aIMG, 0, aIMG.h-1, aIMG.w-1, aIMG.h-1, border_color);
  PutFilledRect(&aIMG, 0, 0, 0, aIMG.h-1, border_color);
  PutFilledRect(&aIMG, aIMG.w-1, 0, aIMG.w-1, aIMG.h-1, border_color);



  //rank
  GetStringSizeFNT(&RankFNT, RankString[rank], &w, &h);
  x = 14-w/2; if (x<0) x = 0;
  y = 0;
  DrawStringFNT(&RankFNT, RankString[rank], &aIMG, x, y, SuitColor[suit]);
  x = aIMG.w-w-x; if (x>aIMG.w-w) x = aIMG.w-w;
  y = aIMG.h-h-y;
  DrawStringRevFlipFNT(&RankFNT, RankString[rank], &aIMG, x, y, SuitColor[suit]);



  //suit
  GetCharSizeFNT(&SuitFNT, SuitChar[suit], &w, &h);
  x = 14-w/2; if (x<0) x = 0;
  y = 20;
  DrawCharFNT(&SuitFNT, SuitChar[suit], &aIMG, x, y, SuitColor[suit]);
  x = aIMG.w-w-x; if (x>aIMG.w-w) x = aIMG.w-w;
  y = aIMG.h-h-y;
  DrawCharRevFlipFNT(&SuitFNT, SuitChar[suit], &aIMG, x, y, SuitColor[suit]);



  if (rank<10)
  {
    if (rank==0 && suit==3)
    {
      //big ace of spades
      MaskedBlit(&SpadeIMG, 0, 0, &aIMG, (aIMG.w-SpadeIMG.w)/2, -40, SpadeIMG.w, SpadeIMG.h);
    }
    else
    {
      //non-court pattern
      GetCharSizeFNT(&SuitPatternFNT, SuitChar[suit], &w, &h);

      adjustx = 0;
      adjusty = -8;
      for (i=0; i<10; i++)
      {
        dx = CardData[2*10*rank+2*i+0]/6.0;
        dy = CardData[2*10*rank+2*i+1]/6.0;
        if (dx==0.0 && dy==0.0) break;
        x = (int)((double)aIMG.w*dx-(double)w*0.5);
        y = (int)((double)aIMG.h*dy-(double)h*0.5);
        if (dy>0.5) {x -= adjustx; y -= adjusty;} else {x += adjustx; y += adjusty;}
        if (rank==0 && suit==3) y -= 18;

        if (dy>0.5) DrawCharRevFlipFNT(&SuitPatternFNT, SuitChar[suit], &aIMG, x, y, SuitColor[suit]);
        else DrawCharFNT(&SuitPatternFNT, SuitChar[suit], &aIMG, x, y, SuitColor[suit]);
      }
    }
  }



  //rescale card by one-half for antialiasing
  Rescale(&aIMG, &bIMG);



  //court image and border
  if (rank>=10)
  {
    w = CourtIMG.w/12;
    h = CourtIMG.h;

    x = (bIMG.w-w)/2;
    y = (bIMG.h-h)/2;

    Blit(&CourtIMG, w*(3*suit+(rank-10)), 0, &bIMG, x, y, w, h);

    PutFilledRect(&bIMG, x-1, y-1, x+w, y-1, BLUE);
    PutFilledRect(&bIMG, x-1, y+h, x+w, y+h, BLUE);
    PutFilledRect(&bIMG, x-1, y-1, x-1, y+h, BLUE);
    PutFilledRect(&bIMG, x+w, y-1, x+w, y+h, BLUE);
  }



  x = CARDW*(index%10);
  y = CARDH*(index/10);
  Blit(&bIMG, 0, 0, &CardsIMG, x, y, CARDW, CARDH);

  CardsRendered[index] = 1;



cleanup:

  FreeIMG(&aIMG);
  FreeIMG(&bIMG);



render:

  if (CardsRendered[index])
  {
    x = CARDW*(index%10);
    y = CARDH*(index/10);
    Blit(&CardsIMG, x, y, &ClientIMG, cx, cy, CARDW, CARDH);
  }
}



void RenderFoundationRank(int cx, int cy)
{
  IMG aIMG = {0, 0, 0};

  int w, h, x, y;
  char *string;

  if (FoundationRankRendered==FoundationRank) goto render;

  if (AllocIMG(&aIMG, CARDW*2, CARDH*2)==0) goto cleanup;

  PutFilledRect(&aIMG, 0, 0, CARDW*2-1, CARDH*2-1, GREY);

  if (FoundationRank==-1) string = "?"; else string = RankString[FoundationRank];

  GetStringSizeFNT(&RankFNT, string, &w, &h);
  x = 14-w/2; if (x<0) x = 0;
  y = 0;
  DrawStringFNT(&RankFNT, string, &aIMG, x, y, DGREY);
  x = aIMG.w-w-x; if (x>aIMG.w-w) x = aIMG.w-w;
  y = aIMG.h-h-y;
  DrawStringRevFlipFNT(&RankFNT, string, &aIMG, x, y, DGREY);

  Rescale(&aIMG, &FoundationRankIMG);

  FoundationRankRendered = FoundationRank;



cleanup:

  FreeIMG(&aIMG);



render:

  if (FoundationRankRendered==FoundationRank)
  {
    Blit(&FoundationRankIMG, 0, 0, &ClientIMG, cx, cy, CARDW, CARDH);
  }
}
//****************************************************************************



//***************************************************************************
//----------
//Easter egg
//----------



#define EASTEREGG_W  40
#define EASTEREGG_H  30



//drop a card in lower-left, upper-left, upper-right, lower-right corners of window, in that order, to activate easter egg
void DetectEasterEgg(int mx, int my)
{
  int x1, x2, y1, y2;

  x1 = 0+EASTEREGG_W; x2 = CLIENTW-1-EASTEREGG_W;
  y1 = 0+EASTEREGG_H; y2 = CLIENTH-1-EASTEREGG_H;

       if (EasterEggCount==0 && mx<=x1 && my>=y2) EasterEggCount++;
  else if (EasterEggCount==1 && mx<=x1 && my<=y1) EasterEggCount++;
  else if (EasterEggCount==2 && mx>=x2 && my<=y1) EasterEggCount++;
  else if (EasterEggCount==3 && mx>=x2 && my>=y2)
  {
    EasterEggCount = 0;
    EasterEggFlag = 1-EasterEggFlag;

    memset(CardsRendered, 0, 13*4); //all cards must be (re)rendered
  }
  else EasterEggCount = 0;
}
//***************************************************************************



//***************************************************************************
//--------------------
//Stack identification
//--------------------



int IsWaste(STACK *stack)
{
  if (stack==&Waste) return 1;
  return 0;
}



int IsReserve(STACK *stack)
{
  if (stack==&Reserve) return 1;
  return 0;
}



int IsFoundation(STACK *stack)
{
  int i;

  for (i=0; i<NumFoundationStacks; i++) if (stack==&Foundation[i]) return 1;
  return 0;
}



int IsTableau(STACK *stack)
{
  int i;

  for (i=0; i<NumTableauStacks; i++) if (stack==&Tableau[i]) return 1;
  return 0;
}



int IsCell(STACK *stack)
{
  int i;

  for (i=0; i<NumCellStacks; i++) if (stack==&Cell[i]) return 1;
  return 0;
}
//***************************************************************************



//***************************************************************************
//-------------
//Card movement
//-------------



//if src_card is 0, last card in src_stack is moved
void MoveCardsRoutine(STACK *src_stack, CARD *src_card, STACK *dst_stack)
{
  CARD *s, *sprev, *d;
  char facedown;

  s = src_stack->card;
  if (s==0) return; //no cards in src_stack
  sprev = 0; while (s!=src_card && s->next!=0) {sprev = s; s = s->next;}
  if (sprev==0) src_stack->card = 0; //src_stack is now empty
  else sprev->next = 0; //unlink from src_stack

  d = dst_stack->card;
  if (d==0) dst_stack->card = s; //dst_stack was empty
  else
  {
    //link to end of dst_stack
    while (d->next!=0) d = d->next;
    d->next = s;
  }



  facedown = 0; //default, turn face up
  if (src_stack==&Deck) facedown = 0; //leaving deck, turn face up
  if (dst_stack==&Deck) facedown = 1; //entering deck, turn face down

  //set facedown flag of all moved cards
  while (s)
  {
    s->facedown = facedown;
    s = s->next;
  }
}



void MoveCards(STACK *src_stack, CARD *src_card, STACK *dst_stack)
{
  CARD *c;

  if (src_card==&Recycle) //this special card signals a recycle
  {
    while (src_stack->card) MoveCardsRoutine(src_stack, 0, dst_stack);
  }
  else if (src_card==&FlipLast) //this special card signals a flip of the last card in a stack
  {
    c = dst_stack->card;
    if (c)
    {
      while (c->next) c = c->next;
      if (c->facedown==0) c->facedown = 1; else c->facedown = 0;
    }
  }
  else
  {
    MoveCardsRoutine(src_stack, src_card, dst_stack);
  }
}
//***************************************************************************



//***************************************************************************
//----
//Undo
//----



void ClearUndoEntries(void)
{
  NumUndo = 0;
  UndoIndex = -1;

  UndoRecycleFlag = 0;
}



void ResetUndoCounter(void)
{
  UndoCounter = 0;
}



void AddUndoEntry(STACK *src_stack, CARD *card, STACK *dst_stack)
{
  int i;

  NumUndo = UndoIndex+1;

  if (NumUndo==MAX_UNDO)
  {
    NumUndo--;
    UndoIndex--;

    for (i=0; i<NumUndo; i++) Undo[i] = Undo[i+1];
  }

  Undo[NumUndo].recycle   = UndoRecycleFlag; UndoRecycleFlag = 0;
  Undo[NumUndo].src_stack = src_stack;
  Undo[NumUndo].card      = card;
  Undo[NumUndo].dst_stack = dst_stack;
  Undo[NumUndo].counter   = UndoCounter;

  NumUndo++;
  UndoIndex++;

  UndoCounter++;
}



void UndoEntry(void)
{
  if (Hand.card) return;
  if (UndoIndex<0) return;

again:
  if (Undo[UndoIndex].recycle) NumRecyclesLeft++;
  MoveCards(Undo[UndoIndex].dst_stack, Undo[UndoIndex].card, Undo[UndoIndex].src_stack);
  UndoIndex--;

  if (UndoIndex>=0) if (Undo[UndoIndex+1].counter==Undo[UndoIndex].counter+1) goto again;
}



void RedoEntry(void)
{
  if (Hand.card) return;
  if (UndoIndex>=NumUndo-1) return;

again:
  UndoIndex++;
  if (Undo[UndoIndex].recycle) NumRecyclesLeft--;
  MoveCards(Undo[UndoIndex].src_stack, Undo[UndoIndex].card, Undo[UndoIndex].dst_stack);

  if (UndoIndex<NumUndo-1) if (Undo[UndoIndex+1].counter==Undo[UndoIndex].counter+1) goto again;
}
//***************************************************************************



//***************************************************************************
//----------
//Detect win
//----------



int IsWon_Klondike(void)
{
  int i;

  if (Hand.card) return 0;
  if (Deck.card) return 0;
  if (Waste.card) return 0;
  for (i=0; i<NumTableauStacks; i++) if (Tableau[i].card) return 0;

  return 1;
}



int IsWon_FreeCell(void)
{
  int i;

  if (Hand.card) return 0;
  for (i=0; i<NumTableauStacks; i++) if (Tableau[i].card) return 0;
  for (i=0; i<NumCellStacks; i++) if (Cell[i].card) return 0;

  return 1;
}



int IsWon_Canfield(void)
{
  int i;

  if (Hand.card) return 0;
  if (Deck.card) return 0;
  if (Waste.card) return 0;
  if (Reserve.card) return 0;
  for (i=0; i<NumTableauStacks; i++) if (Tableau[i].card) return 0;

  return 1;
}



int IsWon_Golf(void)
{
  int i;

  if (Hand.card) return 0;
  for (i=0; i<NumTableauStacks; i++) if (Tableau[i].card) return 0;

  return 1;
}



int IsWon_Spider(void)
{
  int i;

  if (Hand.card) return 0;
  if (Deck.card) return 0;
  for (i=0; i<NumTableauStacks; i++) if (Tableau[i].card) return 0;

  return 1;
}



int IsWon_Pyramid(void)
{
  int i;

  if (Hand.card) return 0;
  if (Deck.card) return 0;
  if (Waste.card) return 0;
  for (i=0; i<NumTableauStacks; i++) if (Tableau[i].card) return 0;
  for (i=0; i<NumCellStacks; i++) if (Cell[i].card) return 0;

  return 1;
}



int IsWon_Yukon(void)
{
  int i;

  if (Hand.card) return 0;
  for (i=0; i<NumTableauStacks; i++) if (Tableau[i].card) return 0;

  return 1;
}



int IsWon_Scorpion(void)
{
  int i;

  if (Hand.card) return 0;
  if (Deck.card) return 0;
  for (i=0; i<NumTableauStacks; i++) if (Tableau[i].card) return 0;

  return 1;
}



int IsWon_MonteCarlo(void)
{
  int i;

  if (Hand.card) return 0;
  if (Deck.card) return 0;
  for (i=0; i<NumTableauStacks; i++) if (Tableau[i].card) return 0;

  return 1;
}



int IsWon_FortyThieves(void)
{
  int i;

  if (Hand.card) return 0;
  if (Deck.card) return 0;
  if (Waste.card) return 0;
  for (i=0; i<NumTableauStacks; i++) if (Tableau[i].card) return 0;

  return 1;
}



int IsWon(void)
{
  switch (GameType)
  {
    case 0: default: return IsWon_Klondike();
    case 1:          return IsWon_FreeCell();
    case 2:          return IsWon_Canfield();
    case 3:          return IsWon_Golf();
    case 4:          return IsWon_Spider();
    case 5:          return IsWon_Pyramid();
    case 6:          return IsWon_Yukon();
    case 7:          return IsWon_Scorpion();
    case 8:          return IsWon_MonteCarlo();
    case 9:          return IsWon_FortyThieves();
  }
}
//***************************************************************************



//****************************************************************************
//------------
//Game buttons
//------------



#define NUM_BUTTONS  3

#define BUTTON_W  40
#define BUTTON_H  20

#define BUTTON_SPACING  6

#define BUTTON_FONT  &EasterEggMsgFNT



char *ButtonText[NUM_BUTTONS] = {"Menu", "Undo", "Redo"};

int PressedButton = -1;



void GetButtonXY(int i, int *x, int *y)
{
  switch (GameType)
  {
    case 0: default: *x = 214; *y = 10; break;
    case 1:          *x = 340; *y = 10; break;
    case 2:          *x = 214; *y = 10; break;
    case 3:          *x =  10; *y = 10; break;
    case 4:          *x = 100; *y = 10; break;
    case 5:          *x = 260; *y = 10; break;
    case 6:          *x =  10; *y = 10; break;
    case 7:          *x = 100; *y = 10; break;
    case 8:          *x =  10; *y = 10; break;
    case 9:          *x = 168; *y = 10; break;
  }

  *y += (BUTTON_H+BUTTON_SPACING)*i;
}



int GetSelectedButton(int mx, int my)
{
  int i, x, y;

  for (i=0; i<NUM_BUTTONS; i++)
  {
    GetButtonXY(i, &x, &y); if (mx>=x && mx<x+BUTTON_W && my>=y && my<y+BUTTON_H) break;
  }
  if (i==NUM_BUTTONS) i = -1;

  return i;
}



void SetPressedButton(int i)
{
  PressedButton = i;
}



void RenderButtons(int mx, int my)
{
  int pos, x, y, i, w, h;
  PIXEL_TYPE color;

  static int old_pos = -1;

  pos = GetSelectedButton(mx, my);
  if (pos!=old_pos)
  {
    if (old_pos!=-1) {GetButtonXY(old_pos, &x, &y); SetUpdateRegion(x, x+BUTTON_W-1, y, y+BUTTON_H-1);}
    if (    pos!=-1) {GetButtonXY(    pos, &x, &y); SetUpdateRegion(x, x+BUTTON_W-1, y, y+BUTTON_H-1);}
  }
  old_pos = pos;

  for (i=0; i<NUM_BUTTONS; i++)
  {
    GetButtonXY(i, &x, &y);

    PutFilledRect(&ClientIMG, x+1, y+1, x+BUTTON_W-2, y+BUTTON_H-2, GREY);

    if (i==PressedButton && (MouseLatch&1)!=0)
    {
      PutFilledRect(&ClientIMG, x, y, x+BUTTON_W-2, y, DGREY);
      PutFilledRect(&ClientIMG, x, y+1, x, y+BUTTON_H-2, DGREY);
      PutFilledRect(&ClientIMG, x, y+BUTTON_H-1, x+BUTTON_W-1, y+BUTTON_H-1, WHITE);
      PutFilledRect(&ClientIMG, x+BUTTON_W-1, y, x+BUTTON_W-1, y+BUTTON_H-2, WHITE);

      color = BLACK;
    }
    else
    {
      PutFilledRect(&ClientIMG, x, y, x+BUTTON_W-1, y, WHITE);
      PutFilledRect(&ClientIMG, x, y+1, x, y+BUTTON_H-1, WHITE);
      PutFilledRect(&ClientIMG, x+1, y+BUTTON_H-1, x+BUTTON_W-1, y+BUTTON_H-1, DGREY);
      PutFilledRect(&ClientIMG, x+BUTTON_W-1, y+1, x+BUTTON_W-1, y+BUTTON_H-2, DGREY);

      if (i==pos && (MouseLatch&1)==0) color = BLACK; else color = DGREY;
    }

    GetStringSizeFNT(BUTTON_FONT, ButtonText[i], &w, &h);
    DrawStringFNT(BUTTON_FONT, ButtonText[i], &ClientIMG, x+(BUTTON_W-w)/2, y+(BUTTON_H-h)/2, color);
  }
}
//****************************************************************************



//***************************************************************************
//---------
//Rendering
//---------



#define NUM_PATTERN_X  5
#define NUM_PATTERN_Y  4



#define SHADOW_W  6
#define SHADOW_H  6



//winning animation
void RenderWin(void)
{
  int dx, dy, ox, oy, o, i, j, suit, w, h, x, y;

  if (IsWon()==0) return;



  dx = CLIENTW/NUM_PATTERN_X;
  dy = CLIENTH/NUM_PATTERN_Y;

  ox = (int)((double)dx*(SynchroCounter%10000000)/10000000);
  oy = (int)((double)dy*(SynchroCounter%10000000)/10000000);

  o = SynchroCounter/10000000;

  for (j=-1; j<NUM_PATTERN_Y; j++)
  for (i=-1; i<NUM_PATTERN_X; i++)
  {
    suit = (j+1 +3-o)&3;

    GetCharSizeFNT(&LargeSuitFNT, SuitChar[suit], &w, &h);

    x = ox+dx/2+dx*i-w/2;
    y = oy+dy/2+dy*j-h/2;

    DrawCharFNT(&LargeSuitFNT, SuitChar[suit], &ClientIMG, x, y, (suit==1 || suit==2)?RED:BLACK);
  }
}



void SpacingRoutine(int type, int facedown, int count, int *x, int *y)
{
  if (type>1) {*x += 18; *y += 0;}
  else
  {
    if (facedown) {*x += 0; *y += 7;}
    else
    {
      if (count>16) {*x += 0; *y += 26/2;}
      else {*x += 0; *y += 26;}
    }
  }
}



//returns 0 if stack is not visible
int GetStackWH(STACK *stack, int *w, int *h)
{
  CARD *c;
  int x, y, count, skipbefore, i;

  x = y = 0;

  if (stack->card==0)
  {
    if (stack->showempty) goto done; else {*w = *h = 0; return 0;}
  }

  c = stack->card; count = 0; while (c) {count++; c = c->next;}
  skipbefore = 0; if (stack->type>0) skipbefore = count-stack->type;

  c = stack->card; i = 0;
  while (c->next)
  {
    if (i>=skipbefore) SpacingRoutine(stack->type, c->facedown, count, &x, &y);
    i++; c = c->next;
  }

done:
  *w = x+CARDW;
  *h = y+CARDH;
  return 1;
}



void RenderStack(STACK *stack)
{
  CARD *c;
  int x, y, count, skipbefore, i;

  x = stack->x; y = stack->y;

  if (stack->card==0)
  {
    if (stack->showempty)
    {
      if (IsFoundation(stack)) RenderFoundationRank(x, y); else PutFilledRect(&ClientIMG, x, y, x+CARDW-1, y+CARDH-1, GREY);
    }
    return;
  }

  c = stack->card; count = 0; while (c) {count++; c = c->next;}
  skipbefore = 0; if (stack->type>0) skipbefore = count-stack->type;

  c = stack->card; i = 0;
  while (c)
  {
    if (i>=skipbefore)
    {
      if (c->facedown) Blit(&BackIMG, 0, 0, &ClientIMG, x, y, CARDW, CARDH);
      else RenderCard(c->rank, c->suit, x, y);

      SpacingRoutine(stack->type, c->facedown, count, &x, &y);
    }
    i++; c = c->next;
  }
}



void RenderTable(void)
{

#if PIXEL_BITS==16
  ClearIMG(&ClientIMG, COLOR(46, 123, 194));
#endif

#if PIXEL_BITS==32
  GradientFillClientIMG(COLOR(46, 123, 194), COLOR(19, 48, 75));
#endif

}



void Render(int mx, int my)
{
  int i, w, h;
  char string[128];

  RenderTable();

  RenderWin();

  RenderButtons(mx, my);

  //stacks on table
  RenderStack(&Deck);
  RenderStack(&Waste);
  RenderStack(&Reserve);
  for (i=0; i<NumFoundationStacks; i++) RenderStack(&Foundation[i]);
  for (i=0; i<NumTableauStacks; i++) RenderStack(&Tableau[i]);
  for (i=0; i<NumCellStacks; i++) RenderStack(&Cell[i]);

  //hand and shadow
  if (GetStackWH(&Hand, &w, &h))
  {
    Hand.x = mx+HandOffsetX;
    Hand.y = my+HandOffsetY;
    RenderStack(&Hand);
    ShadeRect(&ClientIMG, Hand.x+w, Hand.y+SHADOW_H, Hand.x+w+SHADOW_W-1, Hand.y+SHADOW_H+h-1);
    ShadeRect(&ClientIMG, Hand.x+SHADOW_W, Hand.y+h, Hand.x+w-1, Hand.y+h+SHADOW_H-1);
  }

  //game title
  GetStringSizeFNT(&TextFNT, GameTitle[GameType], &w, &h);
  DrawStringFNT(&TextFNT, GameTitle[GameType], &ClientIMG, (0+4)+2, (CLIENTH-h-22)+2, BLACK);
  DrawStringFNT(&TextFNT, GameTitle[GameType], &ClientIMG, (0+4)+0, (CLIENTH-h-22)+0, WHITE);

  //game seed
  sprintf(string, "%u", GameSeed);
  GetStringSizeFNT(&TextFNT, string, &w, &h);
  DrawStringFNT(&TextFNT, string, &ClientIMG, (CLIENTW-w-4)+2, (CLIENTH-h-22)+2, BLACK);
  DrawStringFNT(&TextFNT, string, &ClientIMG, (CLIENTW-w-4)+0, (CLIENTH-h-22)+0, WHITE);

  //foreground message
  if (ForegroundMessage[0])
  {
    GetStringSizeFNT(&TextFNT, ForegroundMessage, &w, &h);
    DrawStringFNT(&TextFNT, ForegroundMessage, &ClientIMG, (CLIENTW-w)/2+2, (CLIENTH-h-22)+2, BLACK);
    DrawStringFNT(&TextFNT, ForegroundMessage, &ClientIMG, (CLIENTW-w)/2+0, (CLIENTH-h-22)+0, WHITE);
  }
}
//***************************************************************************



//***************************************************************************
//-------------------------
//Pointer utility functions
//-------------------------



#define NEAR_X  (CARDW/2)
#define NEAR_Y  (CARDH/2)



#define ABS(x)  ( (x)<0?-(x):(x) )



int IsNear(int drop_flag, int x, int y, int mx, int my)
{
  if (drop_flag)
  {
    if (ABS(mx+HandOffsetX-x)<=NEAR_X && ABS(my+HandOffsetY-y)<=NEAR_Y) return 1;
    return 0;
  }
  else
  {
    if (mx>=x && mx<x+CARDW && my>=y && my<y+CARDH) return 1;
    return 0;
  }
}



CARD *GetCardUnderPointer(int drop_flag, STACK *stack, int *cx, int *cy, int mx, int my)
{
  CARD *c, *found;
  int count, skipbefore, i, x, y;

  if (stack->card==0) return 0;

  c = stack->card; count = 0; while (c) {count++; c = c->next;}
  skipbefore = 0; if (stack->type>0) skipbefore = count-stack->type;

  c = stack->card; i = 0; found = 0; x = stack->x; y = stack->y;
  while (c)
  {
    if (i>=skipbefore)
    {
      if (IsNear(drop_flag, x, y, mx, my)) {found = c; *cx = x; *cy = y;}

      SpacingRoutine(stack->type, c->facedown, count, &x, &y);
    }
    i++; c = c->next;
  }
  return found;
}



int IsEmptyStackUnderPointer(int drop_flag, STACK *stack, int *x, int *y, int mx, int my)
{
  *x = stack->x;
  *y = stack->y;
  if (stack->showempty && IsNear(drop_flag, *x, *y, mx, my)) return 1;
  return 0;
}



#define GET_STACK_AND_CARD_ROUTINE \
  c = GetCardUnderPointer(drop_flag, *stack, x, y, mx, my); if (c) return c; \
  if (IsEmptyStackUnderPointer(drop_flag, *stack, x, y, mx, my)) return 0;



CARD *GetStackAndCard(int drop_flag, STACK **stack, int *x, int *y, int mx, int my)
{
  CARD *c;
  int i;

  *stack = &Deck   ; GET_STACK_AND_CARD_ROUTINE
  *stack = &Waste  ; GET_STACK_AND_CARD_ROUTINE
  *stack = &Reserve; GET_STACK_AND_CARD_ROUTINE

  for (i=NumFoundationStacks-1; i>=0; i--) {*stack = &Foundation[i]; GET_STACK_AND_CARD_ROUTINE}
  for (i=NumTableauStacks   -1; i>=0; i--) {*stack =    &Tableau[i]; GET_STACK_AND_CARD_ROUTINE}
  for (i=NumCellStacks      -1; i>=0; i--) {*stack =       &Cell[i]; GET_STACK_AND_CARD_ROUTINE}

  *stack = 0;
  return 0;
}
//***************************************************************************



//***************************************************************************
//--------------------------------
//Stack and card utility functions
//--------------------------------



CARD *GetLastCard(STACK *stack)
{
  CARD *c;

  c = stack->card;
  if (c==0) return 0;
  while (c->next) c = c->next;
  return c;
}



int IsEmpty(STACK *stack)
{
  if (stack->card==0) return 1;
  return 0;
}



int GetCardCount(STACK *stack)
{
  CARD *c;
  int count;

  c = stack->card;
  count = 0; while (c) {count++; c = c->next;}
  return count;
}



//return: -1 nothing to compare 0 different colors 1 same color
int IsSameColor(CARD *c1, CARD *c2)
{
  int f1, f2;

  if (c1==0 || c2==0) return -1;

  if (c1->suit==0 || c1->suit==3) f1 = 0; else f1 = 1;
  if (c2->suit==0 || c2->suit==3) f2 = 0; else f2 = 1;
  if (f1==f2) return 1;
  return 0;
}



//return: -1 nothing to compare 0 different suits 1 same suit
int IsSameSuit(CARD *c1, CARD *c2)
{
  if (c1==0 || c2==0) return -1;

  if (c1->suit==c2->suit) return 1;
  return 0;
}



//return: -1 nothing to compare 0 not descending 1 descending
int IsDescending(CARD *c1, CARD *c2, int wrap)
{
  int r1, r2;

  if (c1==0 || c2==0) return -1;

  r1 = c1->rank;
  r2 = c2->rank+1; if (wrap) r2 = r2%13;
  if (r1==r2) return 1;
  return 0;
}



//return: -1 nothing to compare 0 not ascending 1 ascending
int IsAscending(CARD *c1, CARD *c2, int wrap)
{
  int r1, r2;

  if (c1==0 || c2==0) return -1;

  r1 = c1->rank+1; if (wrap) r1 = r1%13;
  r2 = c2->rank;
  if (r1==r2) return 1;
  return 0;
}



//if all cards in stack are arranged by descending rank and same suit, return 1
//if no cards, return 0
//if one card, return 1
int IsSequencedDescendingSuit(STACK *stack, int wrap)
{
  CARD *c;

  c = stack->card; if (c==0) return 0;
  while (c->next)
  {
    if (IsDescending(c, c->next, wrap)==0) return 0;
    if (IsSameSuit(c, c->next)==0) return 0;
    c = c->next;
  }
  return 1;
}



//if all cards in stack are arranged by descending rank and alternate color, return 1
//if no cards, return 0
//if one card, return 1
int IsSequencedDescendingAltColor(STACK *stack, int wrap)
{
  CARD *c;

  c = stack->card; if (c==0) return 0;
  while (c->next)
  {
    if (IsDescending(c, c->next, wrap)==0) return 0;
    if (IsSameColor(c, c->next)==1) return 0;
    c = c->next;
  }
  return 1;
}



int IsPyramidStackBlocked(STACK *stack)
{
  int i;

  char index[28] = {1, 3, 4, 6, 7, 8, 10, 11, 12, 13, 15, 16, 17, 18, 19, 21, 22, 23, 24, 25, 26, -1, -1, -1, -1, -1, -1, -1};

  for (i=0; i<NumTableauStacks; i++) if (stack==&Tableau[i]) break;
  if (i<NumTableauStacks)
  {
    i = index[i];
    if (i!=-1)
    {
      if (&Tableau[i+0]==HandFromStack || Tableau[i+0].card!=0) return 1;
      if (&Tableau[i+1]==HandFromStack || Tableau[i+1].card!=0) return 1;
    }
  }
  return 0;
}
//***************************************************************************



//***************************************************************************
//-------------
//Sound routine
//-------------



//wait 1/frac second
void DoDelay(int frac)
{
  int code, ch, mx, my, mb;

  GetElapsed(0);
  while (GetElapsed(1)<1000000*1/frac) GetInput(&code, &ch, &mx, &my, &mb);
}



void SoundRoutine(char *sound)
{
  int code, ch, mx, my, mb;

  PlayAudio(sound);

  GetInput(&code, &ch, &mx, &my, &mb);
  SetUpdateRegion_Full();

  Render(mx, my);

  DoDelay(25); //1/25 second
}
//***************************************************************************



//***************************************************************************
//-------------------
//Game initialization
//-------------------



void InitCards(int num_decks)
{
  int n, i, j, k;

  if (num_decks<1) num_decks = 1;
  if (num_decks>MAX_DECKS) num_decks = MAX_DECKS;

  n = 13*4*num_decks;

  memset(CardsUsed, 0, n);

  seedMT(GameSeed);

  for (i=0; i<n; i++)
  {
    if (i==n-1) j = 0; else j = rangeMT(0, n-1-i);
    for (k=0; k<n; k++) if (CardsUsed[k]==0) if ((j--)==0) break;
    CardsUsed[k] = 1;

    Cards[i].rank = (char)(k%13);
    Cards[i].suit = (char)((k/13)%4);
    Cards[i].facedown = 1;
    if (i==n-1) Cards[i].next = 0; else Cards[i].next = Cards+i+1;
  }
}



//compatible with Microsoft FreeCell
void InitCards_FreeCell(void)
{
  char a[52], b[52];
  int n, i, j;

  n = 52;

  srand(GameSeed);

  for (i=0; i<n; i++) a[i] = (char)i;
  for (i=0; i<n; i++)
  {
    j = rand()%(n-i);
    b[i] = a[j];
    a[j] = a[n-1-i];
  }

  for (i=0; i<n; i++)
  {
    j = b[n-1-i];

    Cards[i].rank = (char)(j/4);
    Cards[i].suit = (char)(j%4);
    Cards[i].facedown = 1;
    if (i==n-1) Cards[i].next = 0; else Cards[i].next = Cards+i+1;
  }
}



void InitGame_Klondike(void)
{
  int i, j;
  CARD *c;

  InitCards(1);

  FoundationRank = 0; //ace
  NumRecyclesLeft = -1; //infinite
  NumCardsDrawn = 3;
  UseWasteAsFoundation = 0;
  DealToEachTableauFromDeck = 0;
  FillCellsBetweenDeckAndWaste = 0;
  UseThirteenElimination = 0;
  UseMatchingRankElimination = 0;
  ConsolidateTableauAndFillFromDeck = 0;

  NumFoundationStacks = 4;
  NumTableauStacks = 7;
  NumCellStacks = 0;

  Hand.card = 0;
  Hand.x = 0;
  Hand.y = 0;
  Hand.type = 0;
  Hand.showempty = 0;

  Deck.card = Cards+0;
  Deck.x = 10;
  Deck.y = 10;
  Deck.type = 1;
  Deck.showempty = 1;

  Waste.card = 0;
  Waste.x = 10+CARDW+10;
  Waste.y = 10;
  Waste.type = 3;
  Waste.showempty = 1;

  //not used (position offscreen)
  Reserve.card = 0;
  Reserve.x = CLIENTW;
  Reserve.y = CLIENTH;
  Reserve.type = 0;
  Reserve.showempty = 0;

  for (i=0; i<NumFoundationStacks; i++)
  {
    Foundation[i].card = 0;
    Foundation[i].x = CLIENTW-(CARDW+10)*NumFoundationStacks+(CARDW+10)*i;
    Foundation[i].y = 10;
    Foundation[i].type = 1;
    Foundation[i].showempty = 1;
  }

  for (i=0; i<NumTableauStacks; i++)
  {
    Tableau[i].card = 0;
    Tableau[i].x = (CLIENTW-(CARDW+10)*NumTableauStacks+10)/2+(CARDW+10)*i;
    Tableau[i].y = 10+CARDH+10;
    Tableau[i].type = 0;
    Tableau[i].showempty = 1;
  }



  //deal to Tableau
  for (i=0; i<NumTableauStacks; i++)
  {
    for (j=0; j<i+1; j++)
    {
      MoveCards(&Deck, 0, &Tableau[i]);

      //set facedown flag on all but last card moved to stack
      if (j<i)
      {
        c = Tableau[i].card; if (c) {while (c->next) c = c->next; c->facedown = 1;}
      }

      SoundRoutine("DealSound");
    }
  }
}



void InitGame_FreeCell(void)
{
  int i;

  InitCards_FreeCell();

  FoundationRank = 0; //ace
  NumRecyclesLeft = -1; //infinite
  NumCardsDrawn = 0; //not used
  UseWasteAsFoundation = 0;
  DealToEachTableauFromDeck = 0;
  FillCellsBetweenDeckAndWaste = 0;
  UseThirteenElimination = 0;
  UseMatchingRankElimination = 0;
  ConsolidateTableauAndFillFromDeck = 0;

  NumFoundationStacks = 4;
  NumTableauStacks = 8;
  NumCellStacks = 4;

  Hand.card = 0;
  Hand.x = 0;
  Hand.y = 0;
  Hand.type = 0;
  Hand.showempty = 0;

  //not used (position offscreen)
  Deck.card = Cards+0;
  Deck.x = CLIENTW;
  Deck.y = CLIENTH;
  Deck.type = 0;
  Deck.showempty = 1;

  //not used (position offscreen)
  Waste.card = 0;
  Waste.x = CLIENTW;
  Waste.y = CLIENTH;
  Waste.type = 0;
  Waste.showempty = 0;

  //not used (position offscreen)
  Reserve.card = 0;
  Reserve.x = CLIENTW;
  Reserve.y = CLIENTH;
  Reserve.type = 0;
  Reserve.showempty = 0;

  for (i=0; i<NumFoundationStacks; i++)
  {
    Foundation[i].card = 0;
    Foundation[i].x = CLIENTW-(CARDW+10)*NumFoundationStacks+(CARDW+10)*i;
    Foundation[i].y = 10;
    Foundation[i].type = 1;
    Foundation[i].showempty = 1;
  }

  for (i=0; i<NumTableauStacks; i++)
  {
    Tableau[i].card = 0;
    Tableau[i].x = (CLIENTW-(CARDW+10)*NumTableauStacks+10)/2+(CARDW+10)*i;
    Tableau[i].y = 10+CARDH+10;
    Tableau[i].type = 0;
    Tableau[i].showempty = 1;
  }

  for (i=0; i<NumCellStacks; i++)
  {
    Cell[i].card = 0;
    Cell[i].x = 10+(CARDW+10)*i;
    Cell[i].y = 10;
    Cell[i].type = 0;
    Cell[i].showempty = 1;
  }



  //deal to Tableau
  i = 0;
  while (Deck.card)
  {
    MoveCards(&Deck, 0, &Tableau[i]);
    i = (i+1)%NumTableauStacks;

    SoundRoutine("DealSound");
  }
}



void InitGame_Canfield(void)
{
  int i;
  CARD *c;

  InitCards(1);

  FoundationRank = -1; //not set
  NumRecyclesLeft = -1; //infinite
  NumCardsDrawn = 3;
  UseWasteAsFoundation = 0;
  DealToEachTableauFromDeck = 0;
  FillCellsBetweenDeckAndWaste = 0;
  UseThirteenElimination = 0;
  UseMatchingRankElimination = 0;
  ConsolidateTableauAndFillFromDeck = 0;

  NumFoundationStacks = 4;
  NumTableauStacks = 4;
  NumCellStacks = 0;

  Hand.card = 0;
  Hand.x = 0;
  Hand.y = 0;
  Hand.type = 0;
  Hand.showempty = 0;

  Deck.card = Cards+0;
  Deck.x = 10;
  Deck.y = 10;
  Deck.type = 1;
  Deck.showempty = 1;

  Waste.card = 0;
  Waste.x = 10+CARDW+10;
  Waste.y = 10;
  Waste.type = 3;
  Waste.showempty = 1;

  Reserve.card = 0;
  Reserve.x = 10+CARDW+10;
  Reserve.y = 10+CARDH+10;
  Reserve.type = 0;
  Reserve.showempty = 1;

  for (i=0; i<NumFoundationStacks; i++)
  {
    Foundation[i].card = 0;
    Foundation[i].x = CLIENTW-(CARDW+10)*NumFoundationStacks+(CARDW+10)*i;
    Foundation[i].y = 10;
    Foundation[i].type = 1;
    Foundation[i].showempty = 1;
  }

  for (i=0; i<NumTableauStacks; i++)
  {
    Tableau[i].card = 0;
    Tableau[i].x = (CLIENTW-(CARDW+10)*NumTableauStacks+10)/2+(CARDW+10)*i;
    Tableau[i].y = 10+CARDH+10;
    Tableau[i].type = 0;
    Tableau[i].showempty = 1;
  }



  //deal to Reserve
  for (i=0; i<13; i++)
  {
    MoveCards(&Deck, 0, &Reserve);

    //set facedown flag on all but last card moved to stack
    if (i<12)
    {
      c = Reserve.card;
      if (c)
      {
        while (c->next) c = c->next;
        c->facedown = 1;
      }
    }

    SoundRoutine("DealSound");
  }

  //deal to Tableau
  for (i=0; i<NumTableauStacks; i++)
  {
    MoveCards(&Deck, 0, &Tableau[i]);

    SoundRoutine("DealSound");
  }
}



void InitGame_Golf(void)
{
  int i, j;

  InitCards(1);

  FoundationRank = -1; //not set
  NumRecyclesLeft = 0; //none
  NumCardsDrawn = 1;
  UseWasteAsFoundation = 1;
  DealToEachTableauFromDeck = 0;
  FillCellsBetweenDeckAndWaste = 0;
  UseThirteenElimination = 0;
  UseMatchingRankElimination = 0;
  ConsolidateTableauAndFillFromDeck = 0;

  NumFoundationStacks = 0;
  NumTableauStacks = 7;
  NumCellStacks = 0;

  Hand.card = 0;
  Hand.x = 0;
  Hand.y = 0;
  Hand.type = 0;
  Hand.showempty = 0;

  Deck.card = Cards+0;
  Deck.x = (CLIENTW-(CARDW+10)*2+10)/2+(CARDW+10)*0;
  Deck.y = (CLIENTH-CARDH)/2;
  Deck.type = 1;
  Deck.showempty = 1;

  Waste.card = 0;
  Waste.x = (CLIENTW-(CARDW+10)*2+10)/2+(CARDW+10)*1;
  Waste.y = (CLIENTH-CARDH)/2;
  Waste.type = 1;
  Waste.showempty = 1;

  //not used (position offscreen)
  Reserve.card = 0;
  Reserve.x = CLIENTW;
  Reserve.y = CLIENTH;
  Reserve.type = 0;
  Reserve.showempty = 0;

  for (i=0; i<NumTableauStacks; i++)
  {
    Tableau[i].card = 0;
    Tableau[i].x = (CLIENTW-(CARDW+10)*NumTableauStacks+10)/2+(CARDW+10)*i;
    Tableau[i].y = 10;
    Tableau[i].type = 0;
    Tableau[i].showempty = 1;
  }



  //deal to Tableau
  for (i=0; i<NumTableauStacks; i++) for (j=0; j<5; j++)
  {
    MoveCards(&Deck, 0, &Tableau[i]);

    SoundRoutine("DealSound");
  }

  //deal to Waste
  MoveCards(&Deck, 0, &Waste);
}



void InitGame_Spider(void)
{
  int n, i, j;
  CARD *c;

  InitCards(2);

  FoundationRank = -1; //not set
  NumRecyclesLeft = 0; //none
  NumCardsDrawn = 0; //not used
  UseWasteAsFoundation = 1;
  DealToEachTableauFromDeck = 1;
  FillCellsBetweenDeckAndWaste = 0;
  UseThirteenElimination = 0;
  UseMatchingRankElimination = 0;
  ConsolidateTableauAndFillFromDeck = 0;

  NumFoundationStacks = 0;
  NumTableauStacks = 10;
  NumCellStacks = 0;

  Hand.card = 0;
  Hand.x = 0;
  Hand.y = 0;
  Hand.type = 0;
  Hand.showempty = 0;

  Deck.card = Cards+0;
  Deck.x = 10;
  Deck.y = 10;
  Deck.type = 1;
  Deck.showempty = 1;

  Waste.card = 0;
  Waste.x = CLIENTW-(CARDW+10);
  Waste.y = 10;
  Waste.type = 1;
  Waste.showempty = 1;

  //not used (position offscreen)
  Reserve.card = 0;
  Reserve.x = CLIENTW;
  Reserve.y = CLIENTH;
  Reserve.type = 0;
  Reserve.showempty = 0;

  for (i=0; i<NumTableauStacks; i++)
  {
    Tableau[i].card = 0;
    Tableau[i].x = (CLIENTW-(CARDW+10)*NumTableauStacks+10)/2+(CARDW+10)*i;
    Tableau[i].y = 10+CARDH+10;
    Tableau[i].type = 0;
    Tableau[i].showempty = 1;
  }



  //deal to Tableau
  for (i=0; i<NumTableauStacks; i++)
  {
    n = 5+(i<4?1:0);
    for (j=0; j<n; j++)
    {
      MoveCards(&Deck, 0, &Tableau[i]);

      //set facedown flag on all but last card moved to stack
      if (j<n-1)
      {
        c = Tableau[i].card; if (c) {while (c->next) c = c->next; c->facedown = 1;}
      }

      SoundRoutine("DealSound");
    }
  }
}



void InitGame_Pyramid(void)
{
  int i;

  char px[28] = {0, -1, 1, -2, 0, 2, -3, -1, 1, 3, -4, -2, 0, 2, 4, -5, -3, -1, 1, 3, 5, -6, -4, -2, 0, 2, 4, 6};
  char py[28] = {0,  1, 1,  2, 2, 2,  3,  3, 3, 3,  4,  4, 4, 4, 4,  5,  5,  5, 5, 5, 5,  6,  6,  6, 6, 6, 6, 6};

  InitCards(1);

  FoundationRank = -1; //not set
  NumRecyclesLeft = 2;
  NumCardsDrawn = 1;
  UseWasteAsFoundation = 0;
  DealToEachTableauFromDeck = 0;
  FillCellsBetweenDeckAndWaste = 1;
  UseThirteenElimination = 1;
  UseMatchingRankElimination = 0;
  ConsolidateTableauAndFillFromDeck = 0;

  NumFoundationStacks = 1;
  NumTableauStacks = 28;
  NumCellStacks = 1;

  Hand.card = 0;
  Hand.x = 0;
  Hand.y = 0;
  Hand.type = 0;
  Hand.showempty = 0;

  Deck.card = Cards+0;
  Deck.x = 10;
  Deck.y = 10;
  Deck.type = 1;
  Deck.showempty = 1;

  Waste.card = 0;
  Waste.x = 10+(CARDW+10)*(NumCellStacks+1);
  Waste.y = 10;
  Waste.type = 1;
  Waste.showempty = 1;

  //not used (position offscreen)
  Reserve.card = 0;
  Reserve.x = CLIENTW;
  Reserve.y = CLIENTH;
  Reserve.type = 0;
  Reserve.showempty = 0;

  //used but hidden offscreen
  for (i=0; i<NumFoundationStacks; i++)
  {
    Foundation[i].card = 0;
    Foundation[i].x = CLIENTW;
    Foundation[i].y = CLIENTH;
    Foundation[i].type = 1;
    Foundation[i].showempty = 1;
  }

  for (i=0; i<NumTableauStacks; i++)
  {
    Tableau[i].card = 0;
    Tableau[i].x = (CLIENTW-CARDW)/2+(CARDW/2+7)*(int)px[i];
    Tableau[i].y = 10+CARDH+10+(CARDH/2-10)*(int)py[i];
    Tableau[i].type = 1;
    Tableau[i].showempty = 0;
  }

  for (i=0; i<NumCellStacks; i++)
  {
    Cell[i].card = 0;
    Cell[i].x = 10+(CARDW+10)*(i+1);
    Cell[i].y = 10;
    Cell[i].type = 1;
    Cell[i].showempty = 1;
  }



  //deal to Tableau
  for (i=0; i<NumTableauStacks; i++)
  {
    MoveCards(&Deck, 0, &Tableau[i]);

    SoundRoutine("DealSound");
  }
}



void InitGame_Yukon(void)
{
  int i, j, n;
  CARD *c;

  InitCards(1);

  FoundationRank = 0; //ace
  NumRecyclesLeft = 0; //not used
  NumCardsDrawn = 0; //not used
  UseWasteAsFoundation = 0;
  DealToEachTableauFromDeck = 0;
  FillCellsBetweenDeckAndWaste = 0;
  UseThirteenElimination = 0;
  UseMatchingRankElimination = 0;
  ConsolidateTableauAndFillFromDeck = 0;

  NumFoundationStacks = 4;
  NumTableauStacks = 7;
  NumCellStacks = 0;

  Hand.card = 0;
  Hand.x = 0;
  Hand.y = 0;
  Hand.type = 0;
  Hand.showempty = 0;

  //not used (position offscreen)
  Deck.card = Cards+0;
  Deck.x = CLIENTW;
  Deck.y = CLIENTH;
  Deck.type = 0;
  Deck.showempty = 0;

  //not used (position offscreen)
  Waste.card = 0;
  Waste.x = CLIENTW;
  Waste.y = CLIENTH;
  Waste.type = 0;
  Waste.showempty = 0;

  //not used (position offscreen)
  Reserve.card = 0;
  Reserve.x = CLIENTW;
  Reserve.y = CLIENTH;
  Reserve.type = 0;
  Reserve.showempty = 0;

  for (i=0; i<NumFoundationStacks; i++)
  {
    Foundation[i].card = 0;
    Foundation[i].x = CLIENTW-(CARDW+10)*NumFoundationStacks+(CARDW+10)*i;
    Foundation[i].y = 10;
    Foundation[i].type = 1;
    Foundation[i].showempty = 1;
  }

  for (i=0; i<NumTableauStacks; i++)
  {
    Tableau[i].card = 0;
    Tableau[i].x = (CLIENTW-(CARDW+10)*NumTableauStacks+10)/2+(CARDW+10)*i;
    Tableau[i].y = 10+CARDH+10;
    Tableau[i].type = 0;
    Tableau[i].showempty = 1;
  }



  //deal to Tableau
  for (i=0; i<NumTableauStacks; i++)
  {
    n = i+(i==0?1:5);
    for (j=0; j<n; j++)
    {
      MoveCards(&Deck, 0, &Tableau[i]);
      if (j<n-(i==0?1:5)) {c = GetLastCard(&Tableau[i]); if (c) c->facedown = 1;}

      SoundRoutine("DealSound");
    }
  }
}



void InitGame_Scorpion(void)
{
  int i, j, n;
  CARD *c;

  InitCards(1);

  FoundationRank = -1; //not set
  NumRecyclesLeft = 0; //none
  NumCardsDrawn = 0; //not used
  UseWasteAsFoundation = 1;
  DealToEachTableauFromDeck = 1;
  FillCellsBetweenDeckAndWaste = 0;
  UseThirteenElimination = 0;
  UseMatchingRankElimination = 0;
  ConsolidateTableauAndFillFromDeck = 0;

  NumFoundationStacks = 0;
  NumTableauStacks = 7;
  NumCellStacks = 0;

  Hand.card = 0;
  Hand.x = 0;
  Hand.y = 0;
  Hand.type = 0;
  Hand.showempty = 0;

  Deck.card = Cards+0;
  Deck.x = 10;
  Deck.y = 10;
  Deck.type = 1;
  Deck.showempty = 1;

  Waste.card = 0;
  Waste.x = CLIENTW-(CARDW+10);
  Waste.y = 10;
  Waste.type = 1;
  Waste.showempty = 1;

  //not used (position offscreen)
  Reserve.card = 0;
  Reserve.x = CLIENTW;
  Reserve.y = CLIENTH;
  Reserve.type = 0;
  Reserve.showempty = 0;

  for (i=0; i<NumTableauStacks; i++)
  {
    Tableau[i].card = 0;
    Tableau[i].x = (CLIENTW-(CARDW+10)*NumTableauStacks+10)/2+(CARDW+10)*i;
    Tableau[i].y = 10+CARDH+10;
    Tableau[i].type = 0;
    Tableau[i].showempty = 1;
  }



  //deal to Tableau
  for (i=0; i<NumTableauStacks; i++)
  {
    n = 7;
    for (j=0; j<n; j++)
    {
      MoveCards(&Deck, 0, &Tableau[i]);
      if (j<n-(i<4?4:7)) {c = GetLastCard(&Tableau[i]); if (c) c->facedown = 1;}

      SoundRoutine("DealSound");
    }
  }
}



void InitGame_MonteCarlo(void)
{
  int i;

  InitCards(1);

  FoundationRank = -1; //not set
  NumRecyclesLeft = 0; //none
  NumCardsDrawn = 0; //not used
  UseWasteAsFoundation = 1;
  DealToEachTableauFromDeck = 0;
  FillCellsBetweenDeckAndWaste = 0;
  UseThirteenElimination = 0;
  UseMatchingRankElimination = 1;
  ConsolidateTableauAndFillFromDeck = 1;

  NumFoundationStacks = 0;
  NumTableauStacks = 25;
  NumCellStacks = 0;

  Hand.card = 0;
  Hand.x = 0;
  Hand.y = 0;
  Hand.type = 0;
  Hand.showempty = 0;

  Deck.card = Cards+0;
  Deck.x = 10;
  Deck.y = (CLIENTH-CARDH)/2;
  Deck.type = 1;
  Deck.showempty = 1;

  //used but hidden offscreen
  Waste.card = 0;
  Waste.x = CLIENTW;
  Waste.y = CLIENTH;
  Waste.type = 1;
  Waste.showempty = 1;

  //not used (position offscreen)
  Reserve.card = 0;
  Reserve.x = CLIENTW;
  Reserve.y = CLIENTH;
  Reserve.type = 0;
  Reserve.showempty = 0;

  for (i=0; i<NumTableauStacks; i++)
  {
    Tableau[i].card = 0;
    Tableau[i].x = (CLIENTW-(CARDW+10)*5+10)/2+(CARDW+10)*(i%5);
    Tableau[i].y = (CLIENTH-(CARDH+10)*5+10)/2+(CARDH+10)*(i/5);
    Tableau[i].type = 1;
    Tableau[i].showempty = 0;
  }



  //deal to Tableau
  for (i=0; i<NumTableauStacks; i++)
  {
    MoveCards(&Deck, 0, &Tableau[i]);

    SoundRoutine("DealSound");
  }
}



void InitGame_FortyThieves(void)
{
  int i, j;

  InitCards(2);

  FoundationRank = 0; //ace
  NumRecyclesLeft = 0; //none
  NumCardsDrawn = 1;
  UseWasteAsFoundation = 0;
  DealToEachTableauFromDeck = 0;
  FillCellsBetweenDeckAndWaste = 0;
  UseThirteenElimination = 0;
  UseMatchingRankElimination = 0;
  ConsolidateTableauAndFillFromDeck = 0;

  NumFoundationStacks = 8;
  NumTableauStacks = 10;
  NumCellStacks = 0;

  Hand.card = 0;
  Hand.x = 0;
  Hand.y = 0;
  Hand.type = 0;
  Hand.showempty = 0;

  Deck.card = Cards+0;
  Deck.x = 10;
  Deck.y = 10;
  Deck.type = 1;
  Deck.showempty = 1;

  Waste.card = 0;
  Waste.x = 10+CARDW+10;
  Waste.y = 10;
  Waste.type = 1;
  Waste.showempty = 1;

  //not used (position offscreen)
  Reserve.card = 0;
  Reserve.x = CLIENTW;
  Reserve.y = CLIENTH;
  Reserve.type = 0;
  Reserve.showempty = 0;

  for (i=0; i<NumFoundationStacks; i++)
  {
    Foundation[i].card = 0;
    Foundation[i].x = CLIENTW-10-(CARDW+2)*NumFoundationStacks+2+(CARDW+2)*i;
    Foundation[i].y = 10;
    Foundation[i].type = 1;
    Foundation[i].showempty = 1;
  }

  for (i=0; i<NumTableauStacks; i++)
  {
    Tableau[i].card = 0;
    Tableau[i].x = (CLIENTW-(CARDW+10)*NumTableauStacks+10)/2+(CARDW+10)*i;
    Tableau[i].y = 10+CARDH+10;
    Tableau[i].type = 0;
    Tableau[i].showempty = 1;
  }



  //deal to Tableau
  for (i=0; i<NumTableauStacks; i++) for (j=0; j<4; j++)
  {
    MoveCards(&Deck, 0, &Tableau[i]);

    SoundRoutine("DealSound");
  }
}



void InitGame(void)
{
  switch (GameType)
  {
    case 0: default: InitGame_Klondike();     break;
    case 1:          InitGame_FreeCell();     break;
    case 2:          InitGame_Canfield();     break;
    case 3:          InitGame_Golf();         break;
    case 4:          InitGame_Spider();       break;
    case 5:          InitGame_Pyramid();      break;
    case 6:          InitGame_Yukon();        break;
    case 7:          InitGame_Scorpion();     break;
    case 8:          InitGame_MonteCarlo();   break;
    case 9:          InitGame_FortyThieves(); break;
  }
}
//***************************************************************************



//***************************************************************************
//----------
//Game logic
//----------



int IsPickupOK(STACK *stack, CARD *card)
{
  if (card->facedown==0) return 1;

  return 0;
}



int IsDropOK_Klondike(STACK *stack)
{
  CARD *c1, *c2;

  if (HandFromStack==stack) return 0;

  c1 = GetLastCard(stack);
  c2 = Hand.card;

  //one card from Waste/Foundation to ...
  if (IsWaste(HandFromStack)==1 || IsFoundation(HandFromStack)==1)
  {
    if (IsTableau(stack)==1)
    {
      if (GetCardCount(&Hand)==1)
      {
        if (IsEmpty(stack)==1)
        {
          //empty Tableau (king)
          if (Hand.card->rank==12) return 1;
        }
        else
        {
          //non-empty Tableau (different colors, unwrapped descending)
          if (IsSameColor(c1, c2)==0) if (IsDescending(c1, c2, 0)==1) return 1;
        }
      }
    }
  }

  //any number of cards from Tableau to ...
  if (IsTableau(HandFromStack)==1)
  {
    if (IsTableau(stack)==1)
    {
      if (IsEmpty(stack)==1)
      {
        //empty Tableau (king)
        if (Hand.card->rank==12) return 1;
      }
      else
      {
        //non-empty Tableau (different colors, unwrapped descending)
        if (IsSameColor(c1, c2)==0) if (IsDescending(c1, c2, 0)==1) return 1;
      }
    }
  }

  //one card from Waste/Tableau to ...
  if (IsWaste(HandFromStack)==1 || IsTableau(HandFromStack)==1)
  {
    if (IsFoundation(stack)==1)
    {
      if (GetCardCount(&Hand)==1)
      {
        if (IsEmpty(stack)==1)
        {
          //empty Foundation
          if (FoundationRank==-1) {FoundationRank = Hand.card->rank; return 1;}
          else if (Hand.card->rank==FoundationRank) return 1;
        }
        else
        {
          //non-empty Foundation (same suit, unwrapped ascending)
          if (IsSameSuit(c1, c2)==1) if (IsAscending(c1, c2, 0)==1) return 1;
        }
      }
    }
  }

  return 0;
}



int IsDropOK_FreeCell(STACK *stack)
{
  CARD *c1, *c2;
  int count, spaces, i;

  if (HandFromStack==stack) return 0;

  c1 = GetLastCard(stack);
  c2 = Hand.card;

  count = GetCardCount(&Hand);

  //more than one card (sequenced descending alt color) from Tableau to ...
  if (IsTableau(HandFromStack)==1 && count>1)
  {
    if (IsTableau(stack)==1)
    {
      if (IsSequencedDescendingAltColor(&Hand, 0)==1)
      {
        spaces = 1;
        for (i=0; i<NumCellStacks; i++) if (Cell[i].card==0) spaces++;
        for (i=0; i<NumTableauStacks; i++) if ((&Tableau[i])!=HandFromStack && (&Tableau[i])!=stack && Tableau[i].card==0) spaces <<= 1;

        if (spaces>=count)
        {
          if (IsEmpty(stack)==1)
          {
            //empty Tableau
            return 1;
          }
          else
          {
            //non-empty Tableau (different colors, unwrapped descending)
            if (IsSameColor(c1, c2)==0) if (IsDescending(c1, c2, 0)==1) return 1;
          }
        }
      }
    }
  }

  //one card from Tableau/Cell to ...
  if ((IsTableau(HandFromStack)==1 || IsCell(HandFromStack)==1) && count==1)
  {
    if (IsTableau(stack)==1)
    {
      if (IsEmpty(stack)==1)
      {
        //empty Tableau
        return 1;
      }
      else
      {
        //non-empty Tableau (different colors, unwrapped descending)
        if (IsSameColor(c1, c2)==0) if (IsDescending(c1, c2, 0)==1) return 1;
      }
    }

    if (IsCell(stack)==1)
    {
      if (IsEmpty(stack)==1)
      {
        //empty Cell
        return 1;
      }
    }

    if (IsFoundation(stack)==1)
    {
      if (IsEmpty(stack)==1)
      {
        //empty Foundation
        if (FoundationRank==-1) {FoundationRank = Hand.card->rank; return 1;}
        else if (Hand.card->rank==FoundationRank) return 1;
      }
      else
      {
        //non-empty Foundation (same suit, unwrapped ascending)
        if (IsSameSuit(c1, c2)==1) if (IsAscending(c1, c2, 0)==1) return 1;
      }
    }
  }

  return 0;
}



int IsDropOK_Canfield(STACK *stack)
{
  CARD *c1, *c2;

  if (HandFromStack==stack) return 0;

  c1 = GetLastCard(stack);
  c2 = Hand.card;

  //one card from Waste/Foundation/Reserve to ...
  if (IsWaste(HandFromStack)==1 || IsFoundation(HandFromStack)==1 || IsReserve(HandFromStack)==1)
  {
    if (IsTableau(stack)==1)
    {
      if (GetCardCount(&Hand)==1)
      {
        if (IsEmpty(stack)==1)
        {
          //empty Tableau
          return 1;
        }
        else
        {
          //non-empty Tableau (different colors, wrapped descending)
          if (IsSameColor(c1, c2)==0) if (IsDescending(c1, c2, 1)==1) return 1;
        }
      }
    }
  }

  //any number of cards from Tableau to ...
  if (IsTableau(HandFromStack)==1)
  {
    if (IsTableau(stack)==1)
    {
      if (IsEmpty(stack)==1)
      {
        //empty Tableau
        return 1;
      }
      else
      {
        //non-empty Tableau (different colors, wrapped descending)
        if (IsSameColor(c1, c2)==0) if (IsDescending(c1, c2, 1)==1) return 1;
      }
    }
  }

  //one card from Waste/Reserve/Tableau to ...
  if (IsWaste(HandFromStack)==1 || IsReserve(HandFromStack)==1 || IsTableau(HandFromStack)==1)
  {
    if (IsFoundation(stack)==1)
    {
      if (GetCardCount(&Hand)==1)
      {
        if (IsEmpty(stack)==1)
        {
          //empty Foundation
          if (FoundationRank==-1) {FoundationRank = Hand.card->rank; return 1;}
          else if (Hand.card->rank==FoundationRank) return 1;
        }
        else
        {
          //non-empty Foundation (same suit, wrapped ascending)
          if (IsSameSuit(c1, c2)==1) if (IsAscending(c1, c2, 1)==1) return 1;
        }
      }
    }
  }

  return 0;
}



int IsDropOK_Golf(STACK *stack)
{
  CARD *c1, *c2;

  if (HandFromStack==stack) return 0;

  c1 = GetLastCard(stack);
  c2 = Hand.card;

  //one card from Tableau to ...
  if (IsTableau(HandFromStack)==1)
  {
    if (IsWaste(stack)==1)
    {
      if (GetCardCount(&Hand)==1)
      {
        if (IsEmpty(stack)==0)
        {
          //non-empty Waste (not onto king, unwrapped ascending/descending)
          if (c1->rank!=12) if (IsAscending(c1, c2, 0)==1 || IsDescending(c1, c2, 0)==1) return 1;
        }
      }
    }
  }

  return 0;
}



int IsDropOK_Spider(STACK *stack)
{
  CARD *c1, *c2;

  if (HandFromStack==stack) return 0;

  c1 = GetLastCard(stack);
  c2 = Hand.card;

  //thirteen cards (sequenced unwrapped descending same suit) from Tableau to Waste
  if (IsTableau(HandFromStack)==1)
  {
    if (IsWaste(stack)==1)
    {
      if (GetCardCount(&Hand)==13)
      {
        if (IsSequencedDescendingSuit(&Hand, 0)==1) return 1;
      }
    }
  }

  //any number of cards (sequenced unwrapped descending same suit) from Tableau to ...
  if (IsTableau(HandFromStack)==1)
  {
    if (IsTableau(stack)==1)
    {
      if (IsSequencedDescendingSuit(&Hand, 0)==1)
      {
        if (IsEmpty(stack)==1)
        {
          //empty Tableau
          return 1;
        }
        else
        {
          //non-empty Tableau (unwrapped descending)
          if (IsDescending(c1, c2, 0)==1) return 1;
        }
      }
    }
  }

  return 0;
}



int IsDropOK_Pyramid(STACK *stack)
{
  CARD *c;

  if (HandFromStack==stack) return 0;

  if (GetCardCount(&Hand)!=1) return 0;

  c = GetLastCard(stack);

  //one card from Cell or Waste or unblocked Tableau to ...
  if (IsCell(HandFromStack)==1 || IsWaste(HandFromStack)==1 || (IsTableau(HandFromStack)==1 && IsPyramidStackBlocked(HandFromStack)==0))
  {
    //Cell or Waste or unblocked Tableau (king, or ranks total 13)
    if (IsCell(stack)==1 || IsWaste(stack)==1 || (IsTableau(stack)==1 && IsPyramidStackBlocked(stack)==0))
    {
      if ((1+Hand.card->rank)==13) return 1;
      if (c) if ((1+Hand.card->rank)+(1+c->rank)==13) return 1;
    }

    //Foundation (king)
    if (IsFoundation(stack)==1)
    {
      if ((1+Hand.card->rank)==13) return 1;
    }
  }

  return 0;
}



int IsDropOK_Yukon(STACK *stack)
{
  CARD *c1, *c2;

  if (HandFromStack==stack) return 0;

  c1 = GetLastCard(stack);
  c2 = Hand.card;

  //any number of cards from Tableau to ...
  if (IsTableau(HandFromStack)==1)
  {
    if (IsTableau(stack)==1)
    {
      if (IsEmpty(stack)==1)
      {
        //empty Tableau (king)
        if (Hand.card->rank==12) return 1;
      }
      else
      {
        //non-empty Tableau (different colors)
        if (IsSameColor(c1, c2)==0) return 1;
      }
    }
  }

  //one card from Tableau to ...
  if (IsTableau(HandFromStack)==1)
  {
    if (IsFoundation(stack)==1)
    {
      if (GetCardCount(&Hand)==1)
      {
        if (IsEmpty(stack)==1)
        {
          //empty Foundation
          if (FoundationRank==-1) {FoundationRank = Hand.card->rank; return 1;}
          else if (Hand.card->rank==FoundationRank) return 1;
        }
        else
        {
          //non-empty Foundation (same suit, unwrapped ascending)
          if (IsSameSuit(c1, c2)==1) if (IsAscending(c1, c2, 0)==1) return 1;
        }
      }
    }
  }

  return 0;
}



int IsDropOK_Scorpion(STACK *stack)
{
  CARD *c1, *c2;

  if (HandFromStack==stack) return 0;

  c1 = GetLastCard(stack);
  c2 = Hand.card;

  //thirteen cards (sequenced unwrapped descending same suit) from Tableau to Waste
  if (IsTableau(HandFromStack)==1)
  {
    if (IsWaste(stack)==1)
    {
      if (GetCardCount(&Hand)==13)
      {
        if (IsSequencedDescendingSuit(&Hand, 0)==1) return 1;
      }
    }
  }

  //any number of cards from Tableau to ...
  if (IsTableau(HandFromStack)==1)
  {
    if (IsTableau(stack)==1)
    {
      if (IsEmpty(stack)==1)
      {
        //empty Tableau (king)
        if (Hand.card->rank==12) return 1;
      }
      else
      {
        //non-empty Tableau (same suit, unwrapped descending)
        if (IsSameSuit(c1, c2)==1) if (IsDescending(c1, c2, 0)==1) return 1;
      }
    }
  }

  return 0;
}



int IsDropOK_MonteCarlo(STACK *stack)
{
  int i, x1, y1, x2, y2;
  CARD *c;

  if (HandFromStack==stack) return 0;

  //get coordinates of Tableau stacks

  for (i=0; i<NumTableauStacks; i++) if (HandFromStack==&Tableau[i]) break;
  x1 = i%5; y1 = i/5;

  for (i=0; i<NumTableauStacks; i++) if (stack==&Tableau[i]) break;
  x2 = i%5; y2 = i/5;

  //one card from Tableau stack to adjacent Tableau stack (same rank)
  if (IsTableau(HandFromStack)==1)
  {
    if (IsTableau(stack)==1)
    {
      if (GetCardCount(&Hand)==1)
      {
        if (x1>=x2-1 && x1<=x2+1 && y1>=y2-1 && y1<=y2+1)
        {
          c = GetLastCard(stack);
          if (c) if (c->rank==Hand.card->rank) return 1;
        }
      }
    }
  }

  return 0;
}



int IsDropOK_FortyThieves(STACK *stack)
{
  CARD *c1, *c2;

  if (HandFromStack==stack) return 0;

  if (GetCardCount(&Hand)!=1) return 0;

  c1 = GetLastCard(stack);
  c2 = Hand.card;

  //one card from Waste/Tableau to ...
  if (IsWaste(HandFromStack)==1 || IsTableau(HandFromStack)==1)
  {
    if (IsTableau(stack)==1)
    {
      if (IsEmpty(stack)==1)
      {
        //empty Tableau
        return 1;
      }
      else
      {
        //non-empty Tableau (same suit, unwrapped descending)
        if (IsSameSuit(c1, c2)==1) if (IsDescending(c1, c2, 0)==1) return 1;
      }
    }

    if (IsFoundation(stack)==1)
    {
      if (IsEmpty(stack)==1)
      {
        //empty Foundation
        if (FoundationRank==-1) {FoundationRank = Hand.card->rank; return 1;}
        else if (Hand.card->rank==FoundationRank) return 1;
      }
      else
      {
        //non-empty Foundation (same suit, unwrapped ascending)
        if (IsSameSuit(c1, c2)==1) if (IsAscending(c1, c2, 0)==1) return 1;
      }
    }
  }

  return 0;
}



int IsDropOK(STACK *stack)
{
  switch (GameType)
  {
    case 0: default: return IsDropOK_Klondike(stack);
    case 1:          return IsDropOK_FreeCell(stack);
    case 2:          return IsDropOK_Canfield(stack);
    case 3:          return IsDropOK_Golf(stack);
    case 4:          return IsDropOK_Spider(stack);
    case 5:          return IsDropOK_Pyramid(stack);
    case 6:          return IsDropOK_Yukon(stack);
    case 7:          return IsDropOK_Scorpion(stack);
    case 8:          return IsDropOK_MonteCarlo(stack);
    case 9:          return IsDropOK_FortyThieves(stack);
  }
}
//***************************************************************************



//***************************************************************************
//----------
//After move
//----------



void AfterMove(void)
{
  int i;
  CARD *c;

  //for each empty Tableau move one card from Reserve
  for (i=0; i<NumTableauStacks; i++) if (Reserve.card!=0 && Tableau[i].card==0)
  {
    AddUndoEntry(&Reserve, 0, &Tableau[i]);
    MoveCards(&Reserve, 0, &Tableau[i]);

    SoundRoutine("DealSound");
  }

  //check each last card in Tableau and flip if face down
  for (i=0; i<NumTableauStacks; i++)
  {
    c = Tableau[i].card;
    if (c)
    {
      while (c->next) c = c->next;
      if (c->facedown)
      {
        AddUndoEntry(&Tableau[i], &FlipLast, &Tableau[i]);
        MoveCards(&Tableau[i], &FlipLast, &Tableau[i]);

        SoundRoutine("FlipSound");
      }
    }
  }

  //check last card in Reserve and flip if face down
  c = Reserve.card;
  if (c)
  {
    while (c->next) c = c->next;
    if (c->facedown)
    {
      AddUndoEntry(&Reserve, &FlipLast, &Reserve);
      MoveCards(&Reserve, &FlipLast, &Reserve);

      SoundRoutine("FlipSound");
    }
  }
}
//***************************************************************************



//***************************************************************************
//--------------------------
//Various gameplay functions
//--------------------------



void MoveFromDeckToCellsAndWaste(void)
{
  int i;
  STACK *from_stack, *to_stack;

  if (FillCellsBetweenDeckAndWaste==0 || NumCellStacks==0)
  {
    if (Deck.card)
    {
      //move one card from Deck to Waste
      AddUndoEntry(&Deck, 0, &Waste);
      MoveCards(&Deck, 0, &Waste);

      SoundRoutine("FlipSound");
    }
    return;
  }

  for (i=NumCellStacks-1; i>=-1; i--)
  {
    if (i==-1) from_stack = &Deck; else from_stack = &Cell[i];
    if (i==NumCellStacks-1) to_stack = &Waste; else to_stack = &Cell[i+1];

    if (from_stack->card)
    {
      //move one card from_stack to_stack
      AddUndoEntry(from_stack, 0, to_stack);
      MoveCards(from_stack, 0, to_stack);

      SoundRoutine("FlipSound");
    }
  }
}



void RecycleFromCellsAndWasteToDeck(void)
{
  int i;

  if (FillCellsBetweenDeckAndWaste)
  {
    for (i=NumCellStacks-1; i>=0; i--)
    {
      if (Cell[i].card)
      {
        //move one card from Cell to Waste
        AddUndoEntry(&Cell[i], 0, &Waste);
        MoveCards(&Cell[i], 0, &Waste);

        SoundRoutine("FlipSound");
      }
    }
  }

  if (Waste.card)
  {
    //recycle all Waste cards back to Deck
    AddUndoEntry(&Waste, &Recycle, &Deck);
    MoveCards(&Waste, &Recycle, &Deck);

    SoundRoutine("RecycleSound");
  }
}



void Function_DealToEachTableauFromDeck(void)
{
  int i;

  for (i=0; i<NumTableauStacks; i++)
  {
    if (Deck.card)
    {
      //move one card from Deck to Tableau
      AddUndoEntry(&Deck, 0, &Tableau[i]);
      MoveCards(&Deck, 0, &Tableau[i]);

      SoundRoutine("DealSound");
    }
  }
}



void Function_ConsolidateTableauAndFillFromDeck(void)
{
  int i, j;

  while (1)
  {
    for (i=0; i<NumTableauStacks; i++) if (Tableau[i].card==0) break;
    if (i==NumTableauStacks) break;

    for (j=i; j<NumTableauStacks; j++) if (Tableau[j].card) break;
    if (j==NumTableauStacks) break;

    //move all cards in one Tableau stack to another Tableau stack
    AddUndoEntry(&Tableau[j], Tableau[j].card, &Tableau[i]);
    MoveCards   (&Tableau[j], Tableau[j].card, &Tableau[i]);

    SoundRoutine("FlipSound");
  }

  for (i=0; i<NumTableauStacks; i++)
  {
    if (Deck.card!=0 && Tableau[i].card==0)
    {
      //move one card from Deck to Tableau stack
      AddUndoEntry(&Deck, 0, &Tableau[i]);
      MoveCards   (&Deck, 0, &Tableau[i]);

      SoundRoutine("DealSound");
    }
  }
}



void ClickedOnDeck(void)
{
  int i;

  if (DealToEachTableauFromDeck) {Function_DealToEachTableauFromDeck(); return;}

  if (ConsolidateTableauAndFillFromDeck) {Function_ConsolidateTableauAndFillFromDeck(); return;}



  if (Deck.card==0)
  {
    if (NumRecyclesLeft==-1 || NumRecyclesLeft>0)
    {
      if (NumRecyclesLeft>0) {NumRecyclesLeft--; UndoRecycleFlag = 1;}

      RecycleFromCellsAndWasteToDeck();

      UndoRecycleFlag = 0; //in case preceding function didn't call AddUndoEntry(), which turns off flag automatically
    }
  }
  else
  {
    for (i=0; i<NumCardsDrawn; i++) MoveFromDeckToCellsAndWaste();
  }
}
//***************************************************************************



//***************************************************************************
//-------------------------------
//Pick up/drop hand and auto move
//-------------------------------



void PickupHand(int mx, int my)
{
  STACK *stack;
  CARD *card;
  int x, y;

  if (Hand.card) return;

  card = GetStackAndCard(0, &stack, &x, &y, mx, my);

  if (stack==&Deck)
  {
    ClickedOnDeck();
  }
  else if (stack!=0 && card!=0)
  {
    if (IsPickupOK(stack, card))
    {
      //pick up card(s) from stack
      MoveCards(stack, card, &Hand);

      HandFromStack = stack;
      HandOffsetX = x-mx;
      HandOffsetY = y-my;

      PlayAudio("PickupSound");
    }
  }
}



void DropHandRoutine(STACK *stack)
{
  CARD *c;

  if (stack==0)
  {
    //return hand to original stack
    MoveCards(&Hand, Hand.card, HandFromStack);

    PlayAudio("DropSound");

    return;
  }

  if (UseThirteenElimination)
  {
    if (GetCardCount(&Hand)==1)
    {
      if ((1+Hand.card->rank)==13) stack = &Foundation[0];
      else
      {
        c = GetLastCard(stack);
        if (c)
        {
          if ((1+c->rank)+(1+Hand.card->rank)==13)
          {
            //move one card from stack to first Foundation
            AddUndoEntry(stack, 0, &Foundation[0]);
            MoveCards(stack, 0, &Foundation[0]);

            stack = &Foundation[0];
          }
        }
      }
    }
  }
  else if (UseMatchingRankElimination)
  {
    if (GetCardCount(&Hand)==1)
    {
      c = GetLastCard(stack);
      if (c)
      {
        if (c->rank==Hand.card->rank)
        {
          //move one card from stack to Waste
          AddUndoEntry(stack, 0, &Waste);
          MoveCards(stack, 0, &Waste);

          stack = &Waste;
        }
      }
    }
  }

  //drop hand on selected stack
  AddUndoEntry(HandFromStack, Hand.card, stack);
  MoveCards(&Hand, Hand.card, stack);

  SoundRoutine("DropSound");

  AfterMove();
}



void DropHand(int mx, int my)
{
  STACK *stack;
  CARD *card;
  int x, y;

  if (Hand.card==0) return;

  card = GetStackAndCard(1, &stack, &x, &y, mx, my);

  if (stack) if (IsDropOK(stack)==0) stack = 0;

  DropHandRoutine(stack);



  DetectEasterEgg(mx, my);
}



void AutoMove(int mx, int my)
{
  int i;
  STACK *stack;

  if (Hand.card) return;
  PickupHand(mx, my);
  if (Hand.card==0) return;

  //determine stack to automove to
  stack = 0;
  if (UseWasteAsFoundation)
  {
    //use Waste as Foundation
    if (IsDropOK(&Waste)) stack = &Waste;
  }
  else
  {
    //use normal Foundation
    for (i=0; i<NumFoundationStacks; i++) if (IsDropOK(&Foundation[i])) break;
    if (i<NumFoundationStacks) stack = &Foundation[i];
  }

  DropHandRoutine(stack);
}
//***************************************************************************



//***************************************************************************
//-------------------------------------
//Interpret output from FreeCell solver
//-------------------------------------



void WriteRank(FILE *f, int rank)
{
       if (rank==0) fputc('A', f);
  else if (rank>=1 && rank<=8) fputc(rank-1+'2', f);
  else if (rank==9 ) fputc('T', f);
  else if (rank==10) fputc('J', f);
  else if (rank==11) fputc('Q', f);
  else if (rank==12) fputc('K', f);
}



void WriteSuit(FILE *f, int suit)
{
       if (suit==0) fputc('C', f);
  else if (suit==1) fputc('D', f);
  else if (suit==2) fputc('H', f);
  else if (suit==3) fputc('S', f);
}



void WriteStack(FILE *f, STACK *stack, int onlylast)
{
  CARD *c;

  c = stack->card;
  while (c)
  {
    if (onlylast==0 || c->next==0)
    {
      WriteRank(f, c->rank);
      WriteSuit(f, c->suit);
      fputc(' ', f);
    }

    c = c->next;
  }
}



void WriteLayout(void)
{
  FILE *f;
  int i;

  f = fopen("layout", "w"); if (f==0) return;

  for (i=0; i<NumTableauStacks; i++)
  {
    WriteStack(f, &Tableau[i], 0);
    fputc('\n', f);
  }

  for (i=0; i<NumCellStacks; i++) WriteStack(f, &Cell[i], 0);
  fputc('\n', f);

  for (i=0; i<NumFoundationStacks; i++) WriteStack(f, &Foundation[i], 1);
  fputc('\n', f);

  fclose(f);
}



//---------------------------------------------------------------------------



CARD *FindCard(int rank, int suit, STACK **stack)
{
  int i;
  CARD *c;

  for (i=0; i<NumTableauStacks; i++)
  {
    c = Tableau[i].card;
    while (c)
    {
      if (c->rank==rank && c->suit==suit) {*stack = &Tableau[i]; return c;}
      c = c->next;
    }
  }

  for (i=0; i<NumCellStacks; i++)
  {
    c = Cell[i].card;
    while (c)
    {
      if (c->rank==rank && c->suit==suit) {*stack = &Cell[i]; return c;}
      c = c->next;
    }
  }

  *stack = 0;
  return 0;
}



int GetRank(int ch)
{
  if (ch=='A') return 0;
  else if (ch>='2' && ch<='9') return ch-'2'+1;
  else if (ch=='T') return 9;
  else if (ch=='J') return 10;
  else if (ch=='Q') return 11;
  else if (ch=='K') return 12;

  return -1;
}



int GetSuit(int ch)
{
  if (ch=='C') return 0;
  else if (ch=='D') return 1;
  else if (ch=='H') return 2;
  else if (ch=='S') return 3;

  return -1;
}



int CountNewlines(char *filename)
{
  FILE *f;
  int count, ch;

  f = fopen(filename, "r"); if (f==0) return 0;
  count = 0;
  while (1)
  {
    ch = fgetc(f); if (ch==EOF) break;
    if (ch=='\n') count++;
  }
  fclose(f);
  return count;
}



void ParseSolution(void)
{
  FILE *f;
  int num_steps, current_step;
  int i, from_rank, from_suit, to_rank, to_suit;
  int to_foundation, to_freecell, to_emptypile;
  CARD *from_card, *to_card;
  STACK *from_stack, *to_stack = 0; //set to_stack to zero just to prevent compiler warning
  int code, ch, mx, my, mb;

  num_steps = CountNewlines("win");
  if (num_steps==0)
  {
    sprintf(ForegroundMessage, "No Solution Found     (E)xit");

    GetInput(&code, &ch, &mx, &my, &mb);
    SetUpdateRegion_Full();
    while (1)
    {
      Render(mx, my);
      GetInput(&code, &ch, &mx, &my, &mb);
      if (ch=='e') break;
    }

    return;
  }

  current_step = 0;

  f = fopen("win", "r"); if (f==0) return;

  while (1)
  {
    sprintf(ForegroundMessage, "Solving Step %i of %i     (S)olve     (E)xit", 1+current_step, num_steps);

    GetInput(&code, &ch, &mx, &my, &mb);
    SetUpdateRegion_Full();
    while (1)
    {
      Render(mx, my);
      GetInput(&code, &ch, &mx, &my, &mb);
      if (ch=='s' || ch=='e') break;
    }
    if (ch=='e') break;



    ch = fgetc(f); if (ch==EOF) break;

    from_rank = GetRank(ch); if (from_rank==-1) break;
    ch = fgetc(f); from_suit = GetSuit(ch); if (from_suit==-1) break;
    if (fgetc(f)!=' ') break;

    to_foundation = 0;
    to_freecell = 0;
    to_emptypile = 0;
    to_rank = -1;
    to_suit = -1;

    ch = fgetc(f);
    if (ch=='o')
    {
      if (fgetc(f)!='u') break;
      if (fgetc(f)!='t') break;
      if (fgetc(f)!='\n') break;

      to_foundation = 1;
    }
    else if (ch=='t')
    {
      if (fgetc(f)!='o') break;
      if (fgetc(f)!=' ') break;
    }
    else break;

    if (to_foundation==0)
    {
      ch = fgetc(f);
      if (ch=='t')
      {
        if (fgetc(f)!='e') break;
        if (fgetc(f)!='m') break;
        if (fgetc(f)!='p') break;
        if (fgetc(f)!='\n') break;

        to_freecell = 1;
      }
      else if (ch=='e')
      {
        if (fgetc(f)!='m') break;
        if (fgetc(f)!='p') break;
        if (fgetc(f)!='t') break;
        if (fgetc(f)!='y') break;
        if (fgetc(f)!=' ') break;
        if (fgetc(f)!='p') break;
        if (fgetc(f)!='i') break;
        if (fgetc(f)!='l') break;
        if (fgetc(f)!='e') break;
        if (fgetc(f)!='\n') break;

        to_emptypile = 1;
      }
      else
      {
        to_rank = GetRank(ch); if (to_rank==-1) break;
        ch = fgetc(f); to_suit = GetSuit(ch); if (to_suit==-1) break;
        if (fgetc(f)!=' ') break;
        if (fgetc(f)!='\n') break;
      }
    }



    from_card = FindCard(from_rank, from_suit, &from_stack); if (from_card==0) break;

    if (to_foundation)
    {
      for (i=0; i<NumFoundationStacks; i++)
      {
        to_stack = &Foundation[i];
        to_card = GetLastCard(to_stack);
        if (IsEmpty(to_stack)==1)
        {
          if (from_card->rank==FoundationRank) break;
        }
        else
        {
          if (IsSameSuit(to_card, from_card)==1) if (IsAscending(to_card, from_card, 0)==1) break;
        }
      }
      if (i==NumFoundationStacks) break;
    }
    else if (to_freecell)
    {
      for (i=0; i<NumCellStacks; i++)
      {
        to_stack = &Cell[i];
        to_card = 0;
        if (IsEmpty(to_stack)==1) break;
      }
      if (i==NumCellStacks) break;
    }
    else if (to_emptypile)
    {
      for (i=0; i<NumTableauStacks; i++)
      {
        to_stack = &Tableau[i];
        to_card = 0;
        if (IsEmpty(to_stack)==1) break;
      }
      if (i==NumTableauStacks) break;
    }
    else
    {
      to_card = FindCard(to_rank, to_suit, &to_stack); if (to_card==0) break;
    }



    //move from_card in from_stack to_stack
    AddUndoEntry(from_stack, from_card, to_stack);
    MoveCards   (from_stack, from_card, to_stack);



    current_step++; if (current_step==num_steps) break;
  }

  fclose(f);
}



//---------------------------------------------------------------------------



int DoesFileExist(char *filename)
{
  FILE *f;

  f = fopen(filename, "r"); if (f==0) return 0;
  fclose(f);
  return 1;
}



void SolveFreeCell(void)
{
  int code, ch, mx, my, mb;

  if (DoesFileExist("patsolve/patsolve.exe")==0) return;



  sprintf(ForegroundMessage, "Solving This Layout: Please Wait");

  GetInput(&code, &ch, &mx, &my, &mb);
  SetUpdateRegion_Full();

  Render(mx, my);
  GetInput(&code, &ch, &mx, &my, &mb);



  remove("layout");
  remove("win");



  WriteLayout();

  if (ExecCommand("patsolve/patsolve.exe -f -a -w8 -t4 layout"))
  {
    while (MonitorExec()==0)
    {
      GetInput(&code, &ch, &mx, &my, &mb);
    }

    ParseSolution();
  }



  remove("layout");
  remove("win");



  ForegroundMessage[0] = 0;
}
//***************************************************************************



//***************************************************************************
//-------------
//Menu and game
//-------------



#define NUM_MENUBUTTONS  (NUM_GAME_TYPES+2)

#define MENU_RETURN  (NUM_GAME_TYPES+0)
#define MENU_EXIT    (NUM_GAME_TYPES+1)

#define MENUBUTTON_SPACING  2

#define MENUBUTTON_FONT  &EasterEggMsgFNT



#define INPUT_W  250
#define INPUT_H  40

#define INPUT_X  ((CLIENTW-INPUT_W)/2)
#define INPUT_Y  (CLIENTH-50-INPUT_H)

#define INPUT_MAX_LEN  10



char *TitleText = "Compact Solitaire";

char *MenuButtonText[NUM_MENUBUTTONS-NUM_GAME_TYPES] = {"Return to game", "Exit"};



void GetMenuButtonPXYWH(int i, char **p, int *x, int *y, int *w, int *h)
{
  if (i<NUM_GAME_TYPES) *p = GameTitle[i]; else *p = MenuButtonText[i-NUM_GAME_TYPES];

  GetStringSizeFNT(MENUBUTTON_FONT, *p, w, h);

  *w += 8;
  *h += 4;

  *x = (CLIENTW-*w)/2;
  *y = 100;

  *y += (*h+MENUBUTTON_SPACING)*i;

  if (i>=MENU_RETURN) *y += 25;
  if (i>=MENU_EXIT  ) *y += 25;
}



int GetSelectedMenuButton(int mx, int my)
{
  int i, x, y, w, h;
  char *p;

  for (i=0; i<NUM_MENUBUTTONS; i++)
  {
    GetMenuButtonPXYWH(i, &p, &x, &y, &w, &h); if (mx>=x && mx<x+w && my>=y && my<y+h) break;
  }
  if (i==NUM_MENUBUTTONS) i = -1;

  return i;
}



int DoMenu(void)
{
  char seed_string[128], *p;
  int code, ch, mx, my, mb, pos, x, y, w, h, i;
  PIXEL_TYPE color;

  static int old_pos = -1;

  seedMT((unsigned int)time(0));
  sprintf(seed_string, "%u", randomMT());

  GetInput(&code, &ch, &mx, &my, &mb);
  SetUpdateRegion_Full();

  while (1)
  {
    RenderTable();



    //render menubuttons

    pos = GetSelectedMenuButton(mx, my);
    if (pos!=old_pos)
    {
      if (old_pos!=-1) {GetMenuButtonPXYWH(old_pos, &p, &x, &y, &w, &h); SetUpdateRegion(x, x+w-1, y, y+h-1);}
      if (    pos!=-1) {GetMenuButtonPXYWH(    pos, &p, &x, &y, &w, &h); SetUpdateRegion(x, x+w-1, y, y+h-1);}
    }
    old_pos = pos;

    for (i=0; i<NUM_MENUBUTTONS; i++)
    {
      if (i==MENU_RETURN && GameType==-1) continue; //don't show this menubutton if there is no game in progress

      GetMenuButtonPXYWH(i, &p, &x, &y, &w, &h);

      PutFilledRect(&ClientIMG, x+1, y+1, x+w-2, y+h-2, GREY);

      PutFilledRect(&ClientIMG, x, y, x+w-1, y, WHITE);
      PutFilledRect(&ClientIMG, x, y+1, x, y+h-1, WHITE);
      PutFilledRect(&ClientIMG, x+1, y+h-1, x+w-1, y+h-1, DGREY);
      PutFilledRect(&ClientIMG, x+w-1, y+1, x+w-1, y+h-2, DGREY);

      if (i==pos && (MouseLatch&1)==0) color = BLACK; else color = DGREY;

      DrawStringFNT(MENUBUTTON_FONT, p, &ClientIMG, x+4, y+2, color);
    }



    //render title text
    GetStringSizeFNT(&TextFNT, TitleText, &w, &h);
    x = (CLIENTW-w)/2;
    y = 10;
    DrawStringFNT(&TextFNT, TitleText, &ClientIMG, x+2, y+2, BLACK);
    DrawStringFNT(&TextFNT, TitleText, &ClientIMG, x+0, y+0, WHITE);



    //render seed string
    PutFilledRect(&ClientIMG, INPUT_X+1, INPUT_Y+1, INPUT_X+INPUT_W-2, INPUT_Y+INPUT_H-2, GREY);

    PutFilledRect(&ClientIMG, INPUT_X, INPUT_Y, INPUT_X+INPUT_W-2, INPUT_Y, DGREY);
    PutFilledRect(&ClientIMG, INPUT_X, INPUT_Y+1, INPUT_X, INPUT_Y+INPUT_H-2, DGREY);
    PutFilledRect(&ClientIMG, INPUT_X, INPUT_Y+INPUT_H-1, INPUT_X+INPUT_W-1, INPUT_Y+INPUT_H-1, WHITE);
    PutFilledRect(&ClientIMG, INPUT_X+INPUT_W-1, INPUT_Y, INPUT_X+INPUT_W-1, INPUT_Y+INPUT_H-2, WHITE);

    DrawStringFNT(&FixedFNT, seed_string, &ClientIMG, INPUT_X, INPUT_Y+1, BLACK);

    GetStringSizeFNT(&FixedFNT, seed_string, &w, &h);
    PutFilledRect(&ClientIMG, INPUT_X+w, INPUT_Y+1, INPUT_X+w+3-1, INPUT_Y+INPUT_H-2, BLACK);



    GetInput(&code, &ch, &mx, &my, &mb);



    i = strlen(seed_string);
    if (ch>='0' && ch<='9' && i<INPUT_MAX_LEN)
    {
      SetUpdateRegion(INPUT_X, INPUT_X+INPUT_W-1, INPUT_Y, INPUT_Y+INPUT_H-1);
      seed_string[i++] = (char)ch; seed_string[i] = 0;
    }
    if (ch=='\b' && i>0)
    {
      SetUpdateRegion(INPUT_X, INPUT_X+INPUT_W-1, INPUT_Y, INPUT_Y+INPUT_H-1);
      seed_string[--i] = 0;
    }



    if ((mb&1)==0)
    {
      if (MouseLatch&1) {MouseLatch &= ~1;}
    }
    else
    {
      if ((MouseLatch&1)==0)
      {
        MouseLatch |= 1;

        if (mx>=INPUT_X && mx<INPUT_X+INPUT_W && my>=INPUT_Y && my<INPUT_Y+INPUT_H)
        {
          SetUpdateRegion(INPUT_X, INPUT_X+INPUT_W-1, INPUT_Y, INPUT_Y+INPUT_H-1);

          seedMT((unsigned int)time(0));
          sprintf(seed_string, "%u", randomMT());
        }

        pos = GetSelectedMenuButton(mx, my);
        if (pos>=0 && pos<NUM_GAME_TYPES)
        {
          char *endptr;

          PlayAudio("ButtonSound");
          DoDelay(5); //1/5 second

          GameType = pos;
          GameSeed = strtoul(seed_string, &endptr, 10);

          ClearUndoEntries();
          InitGame();

          break;
        }
        else if (pos==MENU_RETURN)
        {
          if (GameType!=-1)
          {
            PlayAudio("ButtonSound");

            break;
          }
        }
        else if (pos==MENU_EXIT)
        {
          PlayAudio("ButtonSound");
          DoDelay(5); //1/5 second

          return 1;
        }
      }
    }

    if ((mb&8)==0)
    {
      if (MouseLatch&2) {MouseLatch &= ~2;}
    }
    else
    {
      if ((MouseLatch&2)==0) {MouseLatch |= 2;}
    }
  }

  return 0;
}



//change cursor shape to indicate a possible move
void DoCursorFeedback(int mx, int my)
{
  STACK *stack;
  CARD *card;
  int x, y;

  if (Hand.card)
  {
    card = GetStackAndCard(1, &stack, &x, &y, mx, my);
    if (stack && IsDropOK(stack)) ChangeCursor(1); else ChangeCursor(0);
  }
}



void Game(void)
{
  int code, ch, mx, my, mb, old_mx, old_my, i;

  MouseLatch = 0;
  SynchroCounter = 0;
  GameType = -1;

Menu:
  if (DoMenu()) return;



  GetInput(&code, &ch, &mx, &my, &mb);
  SetUpdateRegion_Full();

  while (1)
  {
    GetElapsed(0);

    if (IsWon()) SetUpdateRegion_Full();

    Render(mx, my);

    old_mx = mx; old_my = my;
    GetInput(&code, &ch, &mx, &my, &mb);

    SynchroCounter += GetElapsed(1);
    SynchroCounter %= 10000000*4;



    ResetUndoCounter();



    if ((mb&1)==0)
    {
      if (MouseLatch&1)
      {
        MouseLatch &= ~1;

        DropHand(mx, my);
        ChangeCursor(-1);

        SetUpdateRegion_Full();
      }
    }
    else
    {
      if ((MouseLatch&1)==0)
      {
        MouseLatch |= 1;

        i = GetSelectedButton(mx, my);
        SetPressedButton(i);
        switch (i)
        {
          case 0: PlayAudio("ButtonSound"); SetPressedButton(-1); goto Menu;

          case 1: PlayAudio("ButtonSound"); UndoEntry(); break;
          case 2: PlayAudio("ButtonSound"); RedoEntry(); break;

          default: PickupHand(mx, my); break;
        }

        SetUpdateRegion_Full();
      }
      else
      {
        //dragging

        int w, h;

        DoCursorFeedback(mx, my);

        if (GetStackWH(&Hand, &w, &h))
        {
          SetUpdateRegion(old_mx+HandOffsetX, old_mx+HandOffsetX+w+SHADOW_W-1, old_my+HandOffsetY, old_my+HandOffsetY+h+SHADOW_H-1);
          SetUpdateRegion(    mx+HandOffsetX,     mx+HandOffsetX+w+SHADOW_W-1,     my+HandOffsetY,     my+HandOffsetY+h+SHADOW_H-1);
        }
      }
    }

    if ((mb&8)==0)
    {
      if (MouseLatch&2) {MouseLatch &= ~2;}
    }
    else
    {
      if ((MouseLatch&2)==0)
      {
        MouseLatch |= 2;

        AutoMove(mx, my);

        SetUpdateRegion_Full();
      }
      else
      {
        //dragging
      }
    }



    if (MouseLatch==0)
    {
      if (ch=='m') goto Menu;

      if (ch=='u') {UndoEntry(); SetUpdateRegion_Full();}
      if (ch=='r') {RedoEntry(); SetUpdateRegion_Full();}

      if (ch=='s' && GameType==1 && Hand.card==0) {SolveFreeCell(); SetUpdateRegion_Full();}
    }
  }
}
//***************************************************************************





//###########################################################################
//--------------------------------------------
//Image initialization and program entry point
//--------------------------------------------



void FreeIMGs(void)
{
  FreeIMG(&BackIMG);
  FreeIMG(&CourtIMG);
  FreeIMG(&CardsIMG);
  FreeIMG(&FoundationRankIMG);
  FreeIMG(&SpadeIMG);
}



void InitIMGs(void)
{
  int w, h;

  atexit(FreeIMGs);



  if (LoadIMG(&BackIMG, "data/back.bmp")==0) exit(1);
  if (LoadIMG(&CourtIMG, "data/court.bmp")==0) exit(1);

  if (AllocIMG(&CardsIMG, CARDW*10, CARDH*6)==0) exit(1);
  if (AllocIMG(&FoundationRankIMG, CARDW, CARDH)==0) exit(1);



  GetCharSizeFNT(&LargeSuitFNT, 0xaa, &w, &h); //spade
  if (AllocIMG(&SpadeIMG, w, h)==0) exit(1);
  ClearIMG(&SpadeIMG, BLACK);
  DrawCharFNT(&LargeSuitFNT, 0xaa, &SpadeIMG, 0, 0, COLOR(8,8,8)); //spade
}



#include <windows.h>



int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
  InitInterface();

  ClearIMG(&ClientIMG, BLACK);

  LoadFonts();

  InitIMGs();

  memset(CardsRendered, 0, 13*4); //all cards must be (re)rendered
  FoundationRankRendered = -2; //foundation rank must be (re)rendered

  Game();

  return 0;
}
//###########################################################################
