//Compact Solitaire
//Version 1.1 (November 2006)
//Copyright 2006 by Donnie Russell



#include <stdio.h>
#include <stdlib.h>

#include "interface.h"



#define FONT_DATA_FILE  "data/font.dat"



FNT EasterEggMsgFNT = {0, 0, 0, 0, {0}, {0}, 0};
FNT RankFNT         = {0, 0, 0, 0, {0}, {0}, 0};
FNT SuitFNT         = {0, 0, 0, 0, {0}, {0}, 0};
FNT SuitPatternFNT  = {0, 0, 0, 0, {0}, {0}, 0};
FNT LargeSuitFNT    = {0, 0, 0, 0, {0}, {0}, 0};
FNT TextFNT         = {0, 0, 0, 0, {0}, {0}, 0};
FNT FixedFNT        = {0, 0, 0, 0, {0}, {0}, 0};



//****************************************************************************
#ifndef BUILD_FONT_DATA



extern int UpdateX1, UpdateX2, UpdateY1, UpdateY2;

extern IMG ClientIMG;



void GetCharSizeFNT(FNT *fnt, int ch, int *w, int *h)
{
  *w = 0;
  *h = 0;

  if (ch<fnt->first_ch || ch>fnt->last_ch) return;
  ch -= fnt->first_ch;

  *w = fnt->w[ch];
  *h = fnt->h[ch];
}



//char is clipped against update region
int DrawCharFNT(FNT *fnt, int ch, IMG *img, int dx, int dy, PIXEL_TYPE color)
{
  int w, h, sx, sy, ss, dd, i, c, count;
  unsigned char *s_base, *s;
  PIXEL_TYPE *d;

  if (ch<fnt->first_ch || ch>fnt->last_ch) return 0;
  ch -= fnt->first_ch;

  w = fnt->w[ch];
  h = fnt->h[ch];

  sx = sy = 0;

  {
    int x1, x2, y1, y2, clip;

    if (img==&ClientIMG) {x1 = UpdateX1; x2 = UpdateX2; y1 = UpdateY1; y2 = UpdateY2;}
    else {x1 = 0; x2 = img->w-1; y1 = 0; y2 = img->h-1;}

    clip = dx-x1; if (clip<0) {w += clip; sx -= clip; dx = x1;}
    clip = dx+w-(x2+1); if (clip>0) w -= clip;
    if (w<=0) goto end;

    clip = dy-y1; if (clip<0) {h += clip; sy -= clip; dy = y1;}
    clip = dy+h-(y2+1); if (clip>0) h -= clip;
    if (h<=0) goto end;
  }



  s_base = fnt->p+(fnt->max_w>>3)*fnt->max_h*ch+(fnt->max_w>>3)*sy+(sx>>3); ss = fnt->max_w>>3;
  d = img->p+img->w*(img->h-1-dy)+dx; dd = img->w+w;

  while (h--)
  {
    s = s_base;
    c = (*s++)<<(sx&7);
    i = w; if (i>8-(sx&7)) i = 8-(sx&7);

    count = w; while (1)
    {
      count -= i;
      while (i--) {if (c&128) *d = color; c <<= 1; d++;}
      if (count) {c = *s++; i = count; if (i>8) i = 8;} else break;
    }

    s_base += ss; d -= dd;
  }



end:
  return fnt->w[ch];
}



//char is not drawn if any part of it is outside update region
int DrawCharRevFlipFNT(FNT *fnt, int ch, IMG *img, int dx, int dy, PIXEL_TYPE color)
{
  int w, h, ss, dd, count, i, c;
  unsigned char *s_base, *s;
  PIXEL_TYPE *d;

  if (ch<fnt->first_ch || ch>fnt->last_ch) return 0;
  ch -= fnt->first_ch;

  w = fnt->w[ch];
  h = fnt->h[ch];

  {
    int x1, x2, y1, y2;

    if (img==&ClientIMG) {x1 = UpdateX1; x2 = UpdateX2; y1 = UpdateY1; y2 = UpdateY2;}
    else {x1 = 0; x2 = img->w-1; y1 = 0; y2 = img->h-1;}

    if (dx<x1 || dx+w-1>x2 || dy<y1 || dy+h-1>y2) goto end;
  }



  s_base = fnt->p+(fnt->max_w>>3)*fnt->max_h*ch; ss = fnt->max_w>>3;
  d = img->p+img->w*(img->h-1-dy   -h+1)+dx   +w-1; dd = img->w+w;

  while (h--)
  {
    s = s_base;

    count = w; while (count)
    {
      i = count; if (i>8) i = 8; count -= i; c = *s++;
      while (i--) {if (c&128) *d = color; c <<= 1; d--;}
    }

    s_base += ss; d += dd;
  }



end:
  return fnt->w[ch];
}



void GetStringSizeFNT(FNT *fnt, char *string, int *w, int *h)
{
  int ch, ww, hh;

  *w = 0;
  *h = 0;

  while (1)
  {
    ch = *string++; if (ch==0) break;

    GetCharSizeFNT(fnt, ch, &ww, &hh);
    *w += ww;
    if (hh>*h) *h = hh;
  }
}



void DrawStringFNT(FNT *fnt, char *string, IMG *img, int x, int y, PIXEL_TYPE color)
{
  int len, i;

  len = strlen(string);
  for (i=0; i<len; i++) x += DrawCharFNT(fnt, string[i], img, x, y, color);
}



void DrawStringRevFlipFNT(FNT *fnt, char *string, IMG *img, int x, int y, PIXEL_TYPE color)
{
  int len, i;

  len = strlen(string);
  for (i=len-1; i>=0; i--) x += DrawCharRevFlipFNT(fnt, string[i], img, x, y, color);
}



#endif
//****************************************************************************



//****************************************************************************
void SaveFNT(FNT *fnt, FILE *f)
{
  int n, i;

  fwrite(&(fnt->first_ch), 2, 1, f);
  fwrite(&(fnt->last_ch ), 2, 1, f);
  fwrite(&(fnt->max_w   ), 2, 1, f);
  fwrite(&(fnt->max_h   ), 2, 1, f);

  n = (fnt->last_ch)-(fnt->first_ch)+1;
  for (i=0; i<n; i++)
  {
    fwrite(&(fnt->w[i]), 2, 1, f);
    fwrite(&(fnt->h[i]), 2, 1, f);
  }

  n = ((fnt->max_w)>>3)*(fnt->max_h)*n;
  fwrite(fnt->p, 1, n, f);
}



int LoadFNT(FNT *fnt, FILE *f)
{
  int n, i;

  fread(&(fnt->first_ch), 2, 1, f);
  fread(&(fnt->last_ch ), 2, 1, f);
  fread(&(fnt->max_w   ), 2, 1, f);
  fread(&(fnt->max_h   ), 2, 1, f);

  n = (fnt->last_ch)-(fnt->first_ch)+1;
  for (i=0; i<n; i++)
  {
    fread(&(fnt->w[i]), 2, 1, f);
    fread(&(fnt->h[i]), 2, 1, f);
  }

  n = ((fnt->max_w)>>3)*(fnt->max_h)*n;
  fnt->p = (unsigned char *)malloc(n); if ((fnt->p)==0) return 0;
  fread(fnt->p, 1, n, f);

  return 1;
}



void FreeFNT(FNT *fnt)
{
  if (fnt->p) {free(fnt->p); fnt->p = 0;}
}
//****************************************************************************



//****************************************************************************
void FreeFonts(void)
{
  FreeFNT(&EasterEggMsgFNT);
  FreeFNT(&RankFNT);
  FreeFNT(&SuitFNT);
  FreeFNT(&SuitPatternFNT);
  FreeFNT(&LargeSuitFNT);
  FreeFNT(&TextFNT);
  FreeFNT(&FixedFNT);
}



void LoadFonts(void)
{
  FILE *f;



  atexit(FreeFonts);



  f = fopen(FONT_DATA_FILE, "rb"); if (f==0) exit(1);

  if (LoadFNT(&EasterEggMsgFNT, f)==0) {fclose(f); exit(1);}
  if (LoadFNT(&RankFNT        , f)==0) {fclose(f); exit(1);}
  if (LoadFNT(&SuitFNT        , f)==0) {fclose(f); exit(1);}
  if (LoadFNT(&SuitPatternFNT , f)==0) {fclose(f); exit(1);}
  if (LoadFNT(&LargeSuitFNT   , f)==0) {fclose(f); exit(1);}
  if (LoadFNT(&TextFNT        , f)==0) {fclose(f); exit(1);}
  if (LoadFNT(&FixedFNT       , f)==0) {fclose(f); exit(1);}

  fclose(f);
}
//****************************************************************************



//****************************************************************************
#ifdef BUILD_FONT_DATA



#include <windows.h>
#include <wingdi.h>



#undef CLIENTW
#undef CLIENTH

#define CLIENTW 300
#define CLIENTH 300



#define MYCLASSNAME  "fntClass"
#define MYTITLE      "fnt"



HINSTANCE hInstance = 0;
HWND hWnd = 0;
HDC hDC = 0;



int NewFNT(FNT *fnt, int first_ch, int last_ch, char *fontname, int fontwidth, int fontheight)
{
  unsigned char c = 0;
  unsigned char *p;
  int num_ch, ch, w, h, max_w, max_h, x, y;
  char text[] = " ";
  HFONT font, oldfont;
  SIZE s;
  COLORREF black = RGB(0, 0, 0);
  int result = 0;

  if (first_ch<0 || first_ch>255 || last_ch<0 || last_ch>255 || first_ch>last_ch) return 0;

  num_ch = last_ch-first_ch+1;

  font = CreateFont(fontheight, fontwidth, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, fontname); if (font==0) return 0;
  oldfont = SelectObject(hDC, font);
  SetBkMode(hDC, OPAQUE);
  SetBkColor(hDC, RGB(0, 0, 0));
  SetTextColor(hDC, RGB(255, 255, 255));

  //find and store the size of each character; find the maximum size of all characters
  max_w = 0;
  max_h = 0;
  for (ch=0; ch<num_ch; ch++)
  {
    text[0] = first_ch+ch; GetTextExtentPoint32(hDC, text, strlen(text), &s);
    w = s.cx; fnt->w[ch] = w; if (w>max_w) max_w = w;
    h = s.cy; fnt->h[ch] = h; if (h>max_h) max_h = h;
  }
  //make maximum width a multiple of 8
  if (max_w&7) max_w += 8-(max_w&7);

  //allocate memory for all character shapes based on maximum size
  p = (unsigned char *)malloc((max_w>>3)*max_h*num_ch); if (p==0) goto done;
  fnt->p = p;

  //render all characters on device context storing shape of each in allocated memory
  for (ch=0; ch<num_ch; ch++)
  {
    text[0] = first_ch+ch; TextOut(hDC, 0, 0, text, strlen(text));
    w = fnt->w[ch];
    h = fnt->h[ch];
    for (y=0; y<max_h; y++) for (x=0; x<max_w; x++)
    {
      if (x<w && y<h && GetPixel(hDC, x, y)!=black) c = (c<<1)|1; else c <<= 1;
      if ((x&7)==7) *p++ = c;
    }
  }

  fnt->first_ch = first_ch;
  fnt->last_ch = last_ch;
  fnt->max_w = max_w;
  fnt->max_h = max_h;

  result = 1;

done:
  SelectObject(hDC, oldfont);
  DeleteObject(font);

  return result;
}



LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  return DefWindowProc(hWnd, uMsg, wParam, lParam);
}



void ShutdownInterface(void)
{
  if (hWnd)
  {
    if (hDC) {ReleaseDC(hWnd, hDC); hDC = 0;}

    DestroyWindow(hWnd); hWnd = 0;
  }

  if (hInstance) {UnregisterClass(MYCLASSNAME, hInstance); hInstance = 0;}
}



void InitInterface(void)
{
  WNDCLASS wc;
  DWORD dwStyle = WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX;
  DWORD dwExStyle = WS_EX_CLIENTEDGE;
  RECT r;
  int ax, ay, bx, by;



  atexit(ShutdownInterface);



  hInstance = GetModuleHandle(0);

  wc.style = CS_HREDRAW|CS_VREDRAW|CS_OWNDC;
  wc.lpfnWndProc = (WNDPROC)WndProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInstance;
  wc.hIcon = LoadIcon(0, IDI_APPLICATION);
  wc.hCursor = LoadCursor(0, IDC_ARROW);
  wc.hbrBackground = 0;
  wc.lpszMenuName = 0;
  wc.lpszClassName = MYCLASSNAME;

  if (!RegisterClass(&wc)) {hInstance = 0; exit(1);}



  hWnd = CreateWindowEx(dwExStyle, MYCLASSNAME, MYTITLE, dwStyle, 0, 0, 0, 0, 0, 0, hInstance, 0);
  if (!hWnd) exit(1);

  hDC = GetDC(hWnd);
  if (!hDC) exit(1);



  r.left = 0; r.right = CLIENTW; r.top = 0; r.bottom = CLIENTH;
  AdjustWindowRectEx(&r, dwStyle, 0, dwExStyle); ax = r.right-r.left; ay = r.bottom-r.top;
  SystemParametersInfo(SPI_GETWORKAREA, 0, &r, 0); bx = r.right-r.left; by = r.bottom-r.top;
  MoveWindow(hWnd, (bx-ax)/2, (by-ay)/2, ax, ay, 1);
  ShowWindow(hWnd, SW_SHOW);
}



int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
  FILE *f;



  InitInterface();



  atexit(FreeFonts);



  if (NewFNT(&EasterEggMsgFNT, ' '   , '~'   , "Arial"      , 0, -16 )==0) exit(1);
  if (NewFNT(&RankFNT        , '0'   , 'Z'   , "Arial"      , 0, -30 )==0) exit(1);
  if (NewFNT(&SuitFNT        , 0xa7  , 0xaa  , "Symbol"     , 0, -30 )==0) exit(1);
  if (NewFNT(&SuitPatternFNT , 0xa7  , 0xaa  , "Symbol"     , 0, -40 )==0) exit(1);
  if (NewFNT(&LargeSuitFNT   , 0xa7  , 0xaa  , "Symbol"     , 0, -200)==0) exit(1);
  if (NewFNT(&TextFNT        , ' '   , '~'   , "Arial"      , 0, -25 )==0) exit(1);
  if (NewFNT(&FixedFNT       , '0'   , '9'   , "Courier New", 0, -40 )==0) exit(1);

  f = fopen(FONT_DATA_FILE, "wb"); if (f==0) exit(1);

  SaveFNT(&EasterEggMsgFNT, f);
  SaveFNT(&RankFNT        , f);
  SaveFNT(&SuitFNT        , f);
  SaveFNT(&SuitPatternFNT , f);
  SaveFNT(&LargeSuitFNT   , f);
  SaveFNT(&TextFNT        , f);
  SaveFNT(&FixedFNT       , f);

  fclose(f);

  return 0;
}



#endif
//****************************************************************************
